import asyncio
import json
import math
import os
import subprocess
from datetime import datetime
from types import TracebackType
from typing import AsyncIterator, Literal, cast

import aiohttp
import numpy as np
import polars as pl
import torch
import wandb
import weave
from openai import AsyncOpenAI
from tqdm import auto as tqdm
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from typing_extensions import Self
from wandb.sdk.wandb_run import Run
from weave.trace.weave_client import WeaveClient

from art.utils.deploy_model import (
    LoRADeploymentJob,
    LoRADeploymentProvider,
    deploy_model,
)
from art.utils.old_benchmarking.calculate_step_metrics import calculate_step_std_dev
from art.utils.output_dirs import (
    get_default_art_path,
    get_model_dir,
    get_output_dir_from_model_properties,
    get_step_checkpoint_dir,
    get_trajectories_split_dir,
)
from art.utils.s3 import (
    ExcludableOption,
    pull_model_from_s3,
    push_model_to_s3,
)
from art.utils.trajectory_logging import serialize_trajectory_groups
from mp_actors import close_proxy, move_to_child_process

from .. import dev
from ..backend import Backend
from ..model import Model, TrainableModel
from ..preprocessing.pack import (
    PackedTensors,
    packed_tensors_from_tokenized_results,
    packed_tensors_to_dir,
    plot_packed_tensors,
)
from ..preprocessing.tokenize import tokenize_trajectory_groups
from ..trajectories import Trajectory, TrajectoryGroup
from ..types import Message, TrainConfig
from ..utils import format_message, get_model_step
from .checkpoints import (
    delete_checkpoints,
)
from .service import ModelService


class LocalBackend(Backend):
    def __init__(self, *, in_process: bool = False, path: str | None = None) -> None:
        """
        Initializes a local, directory-based Backend interface at the given path.

        Note:
            The local Backend uses Weights & Biases for training monitoring.
            If you don't have a W&B account, you can create one at https://wandb.ai.

        Args:
            in_process: Whether to run the local service in-process.
            path: The path to the local directory. Defaults to "{repo_root}/.art".
        """
        self._in_process = in_process
        self._path = path or get_default_art_path()
        os.makedirs(self._path, exist_ok=True)

        # Other initialization
        self._services: dict[str, ModelService] = {}
        self._tokenizers: dict[str, "PreTrainedTokenizerBase"] = {}
        self._wandb_runs: dict[str, Run] = {}
        self._weave_clients: dict[str, WeaveClient] = {}

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self._close()

    async def close(self) -> None:
        """
        If running vLLM in a separate process, this will kill that process and close the communication threads.
        """
        self._close()

    def _close(self) -> None:
        for _, service in self._services.items():
            close_proxy(service)

    async def register(
        self,
        model: Model,
    ) -> None:
        """
        Registers a model with the local Backend for logging and/or training.

        Args:
            model: An art.Model instance.
        """
        output_dir = get_model_dir(model=model, art_path=self._path)
        os.makedirs(output_dir, exist_ok=True)
        with open(f"{output_dir}/model.json", "w") as f:
            json.dump(model.model_dump(), f)

        # Initialize wandb and weave early if this is a trainable model
        if model.trainable and "WANDB_API_KEY" in os.environ:
            _ = self._get_wandb_run(model)

    async def _get_service(self, model: TrainableModel) -> ModelService:
        from ..dev.get_model_config import get_model_config
        from ..torchtune.service import TorchtuneService
        from ..unsloth.decoupled_service import DecoupledUnslothService
        from ..unsloth.service import UnslothService

        if model.name not in self._services:
            config = get_model_config(
                base_model=model.base_model,
                output_dir=get_model_dir(model=model, art_path=self._path),
                config=model._internal_config,
            )
            if config.get("torchtune_args") is not None:
                service_class = TorchtuneService
            elif config.get("_decouple_vllm_and_unsloth", False):
                service_class = DecoupledUnslothService
            else:
                service_class = UnslothService
            self._services[model.name] = service_class(
                model_name=model.name,
                base_model=model.base_model,
                config=config,
                output_dir=get_model_dir(model=model, art_path=self._path),
            )
            if not self._in_process:
                # Kill all "model-service" processes to free up GPU memory
                subprocess.run(["pkill", "-9", "model-service"])
                if isinstance(
                    self._services[model.name],
                    (UnslothService, DecoupledUnslothService),
                ):
                    # To enable sleep mode, import peft before unsloth
                    # Unsloth will issue warnings, but everything appears to be okay
                    if config.get("engine_args", {}).get("enable_sleep_mode", False):
                        os.environ["IMPORT_PEFT"] = "1"
                    # When moving the service to a child process, import unsloth
                    # early to maximize optimizations
                    os.environ["IMPORT_UNSLOTH"] = "1"
                self._services[model.name] = move_to_child_process(
                    self._services[model.name],
                    process_name="model-service",
                )
        return self._services[model.name]

    def _get_packed_tensors(
        self,
        model: TrainableModel,
        trajectory_groups: list[TrajectoryGroup],
        advantage_balance: float,
        allow_training_without_logprobs: bool,
        scale_rewards: bool,
        plot_tensors: bool,
    ) -> PackedTensors | None:
        if model.base_model not in self._tokenizers:
            self._tokenizers[model.base_model] = AutoTokenizer.from_pretrained(
                model.base_model
            )
        tokenizer = self._tokenizers[model.base_model]
        tokenized_results = list(
            tokenize_trajectory_groups(
                tokenizer,
                trajectory_groups,
                allow_training_without_logprobs,
                scale_rewards,
            )
        )
        if not tokenized_results:
            return None
        max_tokens = max(len(result.tokens) for result in tokenized_results)
        # Round up max_tokens to the nearest multiple of 2048
        sequence_length = math.ceil(max_tokens / 2048) * 2048
        # Cap sequence length at the model's max sequence length
        sequence_length = min(
            sequence_length,
            (model._internal_config or dev.InternalModelConfig())
            .get("init_args", {})
            .get("max_seq_length", 32_768),
        )
        packed_tensors = packed_tensors_from_tokenized_results(
            tokenized_results,
            sequence_length,
            pad_token_id=tokenizer.eos_token_id,  # type: ignore
            advantage_balance=advantage_balance,
        )
        if (
            not allow_training_without_logprobs
            and np.isnan(packed_tensors["logprobs"]).all()
        ):
            print(
                "There are no assistant logprobs to train on. Did you forget to include at least one Choice in Trajectory.messages_and_choices?"
            )
            return None
        if plot_tensors:
            plot_packed_tensors(
                packed_tensors, get_model_dir(model=model, art_path=self._path)
            )
        else:
            print(
                f"Packed {len(tokenized_results)} trajectories into {packed_tensors['tokens'].shape[0]} sequences of length {packed_tensors['tokens'].shape[1]}"
            )
        return packed_tensors

    async def _get_step(self, model: TrainableModel) -> int:
        return self.__get_step(model)

    def __get_step(self, model: Model) -> int:
        if model.trainable:
            model = cast(TrainableModel, model)
            return get_model_step(model, self._path)
        # Non-trainable models do not have checkpoints/steps; default to 0
        return 0

    async def _delete_checkpoints(
        self,
        model: TrainableModel,
        benchmark: str,
        benchmark_smoothing: float,
    ) -> None:
        output_dir = get_model_dir(model=model, art_path=self._path)
        # Keep the latest step
        steps_to_keep = [get_model_step(model, self._path)]
        try:
            best_step = (
                pl.read_ndjson(f"{output_dir}/history.jsonl")
                .drop_nulls(subset=[benchmark])
                .group_by("step")
                .mean()
                .with_columns(pl.col(benchmark).ewm_mean(alpha=benchmark_smoothing))
                .sort(benchmark)
                .select(pl.col("step").last())
                .item()
            )
            steps_to_keep.append(best_step)
        except FileNotFoundError:
            print(f'"{output_dir}/history.jsonl" not found')
        except pl.exceptions.ColumnNotFoundError:
            print(f'No "{benchmark}" metric found in history')
        delete_checkpoints(output_dir, steps_to_keep)

    async def _prepare_backend_for_training(
        self,
        model: TrainableModel,
        config: dev.OpenAIServerConfig | None = None,
    ) -> tuple[str, str]:
        service = await self._get_service(model)
        await service.start_openai_server(config=config)
        server_args = (config or {}).get("server_args", {})

        base_url = f"http://{server_args.get('host', '0.0.0.0')}:{server_args.get('port', 8000)}/v1"
        api_key = server_args.get("api_key", None) or "default"

        def done_callback(_: asyncio.Task[None]) -> None:
            close_proxy(self._services.pop(model.name))

        asyncio.create_task(
            self._monitor_openai_server(model.name, base_url, api_key)
        ).add_done_callback(done_callback)

        return base_url, api_key

    async def _monitor_openai_server(
        self, model_name: str, base_url: str, api_key: str
    ) -> None:
        openai_client = AsyncOpenAI(
            base_url=base_url,
            api_key=api_key,
        )
        async with aiohttp.ClientSession() as session:
            while True:
                # Wait 30 seconds before checking again
                await asyncio.sleep(30)
                # If the server is sleeping, skip the check
                if await self._services[model_name].vllm_engine_is_sleeping():
                    continue
                # Check the metrics
                async with session.get(
                    f"{base_url.split('/v1')[0]}/metrics"
                ) as response:
                    metrics = await response.text()
                # Parse Prometheus metrics for running requests
                running_requests = 0
                pending_requests = 0
                for line in metrics.split("\n"):
                    if line.startswith("vllm:num_requests_running"):
                        running_requests = int(float(line.split()[1]))
                    elif line.startswith("vllm:num_requests_waiting"):
                        pending_requests = int(float(line.split()[1]))
                # If there are no running or pending requests, send a health check
                if running_requests == 0 and pending_requests == 0:
                    try:
                        # Send a health check with a 5 second timeout
                        await openai_client.completions.create(
                            model=model_name,
                            prompt="Hi",
                            max_tokens=1,
                            timeout=5,
                        )
                    except Exception as e:
                        # If the server is sleeping, a failed health check is okay
                        if await self._services[model_name].vllm_engine_is_sleeping():
                            continue
                        raise e

    async def _log(
        self,
        model: Model,
        trajectory_groups: list[TrajectoryGroup],
        split: str = "val",
    ) -> None:
        # Save logs for trajectory groups
        parent_dir = get_trajectories_split_dir(
            get_model_dir(model=model, art_path=self._path), split
        )
        os.makedirs(parent_dir, exist_ok=True)

        # Get the file name for the current iteration, or default to 0 for non-trainable models
        iteration = self.__get_step(model)
        file_name = f"{iteration:04d}.jsonl"

        # Write the logs to the file
        with open(f"{parent_dir}/{file_name}", "w") as f:
            f.write(serialize_trajectory_groups(trajectory_groups))

        # Collect all metrics (including reward) across all trajectories
        all_metrics: dict[str, list[float]] = {"reward": [], "exception_rate": []}

        for group in trajectory_groups:
            for trajectory in group:
                if isinstance(trajectory, BaseException):
                    all_metrics["exception_rate"].append(1)
                    continue
                else:
                    all_metrics["exception_rate"].append(0)
                # Add reward metric
                all_metrics["reward"].append(trajectory.reward)

                # Collect other custom metrics
                for metric, value in trajectory.metrics.items():
                    if metric not in all_metrics:
                        all_metrics[metric] = []
                    all_metrics[metric].append(float(value))

        # Calculate averages for all metrics
        averages = {}
        for metric, values in all_metrics.items():
            if len(values) > 0:
                averages[metric] = sum(values) / len(values)

        # Calculate average standard deviation of rewards within groups
        averages["reward_std_dev"] = calculate_step_std_dev(trajectory_groups)

        self._log_metrics(model, averages, split)

    def _trajectory_log(self, trajectory: Trajectory) -> str:
        """Format a trajectory into a readable log string."""
        header = f"reward: {trajectory.reward} {' '.join(f'{k}: {v}' for k, v in trajectory.metrics.items())}\n\n"
        formatted_messages = []
        for message_or_choice in trajectory.messages_and_choices:
            if isinstance(message_or_choice, dict):
                message = message_or_choice
            else:
                message = cast(Message, message_or_choice.message.model_dump())
            formatted_messages.append(format_message(message))
        return header + "\n".join(formatted_messages)

    async def _train_model(
        self,
        model: TrainableModel,
        trajectory_groups: list[TrajectoryGroup],
        config: TrainConfig,
        dev_config: dev.TrainConfig,
        verbose: bool = False,
    ) -> AsyncIterator[dict[str, float]]:
        if verbose:
            print("Starting _train_model")
        service = await self._get_service(model)
        if verbose:
            print("Logging training data to disk...")
        await self._log(model, trajectory_groups, "train")
        if verbose:
            print("Packing tensors...")

        # Count submitted groups and trainable groups
        num_groups_submitted = len(trajectory_groups)
        num_groups_trainable = sum(
            1
            for group in trajectory_groups
            if group and len(set(trajectory.reward for trajectory in group)) > 1
        )

        packed_tensors = self._get_packed_tensors(
            model,
            trajectory_groups,
            advantage_balance=dev_config.get("advantage_balance", 0.0),
            allow_training_without_logprobs=dev_config.get(
                "allow_training_without_logprobs", False
            ),
            scale_rewards=dev_config.get("scale_rewards", True),
            plot_tensors=dev_config.get("plot_tensors", False),
        )
        if packed_tensors is None:
            print(
                "Skipping tuning as there is no suitable data. "
                "This can happen when all the trajectories in the same group "
                "have the same reward and thus no advantage to train on."
            )

            # Still advance the step by renaming the checkpoint directory
            current_step = self.__get_step(model)
            next_step = current_step + 1
            current_checkpoint_dir = get_step_checkpoint_dir(
                get_model_dir(model=model, art_path=self._path), current_step
            )
            next_checkpoint_dir = get_step_checkpoint_dir(
                get_model_dir(model=model, art_path=self._path), next_step
            )

            # If the current checkpoint exists, rename it to the next step
            if os.path.exists(current_checkpoint_dir):
                os.rename(current_checkpoint_dir, next_checkpoint_dir)
                print(
                    f"Advanced step from {current_step} to {next_step} (no training occurred)"
                )

            # Log metrics showing no groups were trainable
            self._log_metrics(
                model,
                {
                    "num_groups_submitted": num_groups_submitted,
                    "num_groups_trainable": 0,
                },
                "train",
                step=next_step,
            )
            return
        disk_packed_tensors = packed_tensors_to_dir(
            packed_tensors, f"{get_model_dir(model=model, art_path=self._path)}/tensors"
        )
        if dev_config.get("scale_learning_rate_by_reward_std_dev", False):
            config = config.model_copy(
                update={
                    "learning_rate": config.learning_rate
                    * self._get_reward_std_dev_learning_rate_multiplier(model)
                }
            )
        results: list[dict[str, float]] = []
        estimated_gradient_steps = disk_packed_tensors["num_sequences"]
        if torchtune_args := (model._internal_config or dev.InternalModelConfig()).get(
            "torchtune_args"
        ):
            tp = torchtune_args.get("tensor_parallel_dim", 1)
            cp = torchtune_args.get("context_parallel_dim", 1)
            world_size = torch.cuda.device_count()
            dp = world_size // (tp * cp)
            estimated_gradient_steps = math.ceil(estimated_gradient_steps / dp)
        pbar = tqdm.tqdm(total=estimated_gradient_steps, desc="train")
        async for result in service.train(
            disk_packed_tensors, config, dev_config, verbose
        ):
            num_gradient_steps = int(
                result.pop("num_gradient_steps", estimated_gradient_steps)
            )
            assert num_gradient_steps == estimated_gradient_steps, (
                f"num_gradient_steps {num_gradient_steps} != estimated_gradient_steps {estimated_gradient_steps}"
            )
            results.append(result)
            yield {**result, "num_gradient_steps": num_gradient_steps}
            pbar.update(1)
            pbar.set_postfix(result)
        pbar.close()
        if verbose:
            print("Logging metrics...")
        data = {
            k: sum(d.get(k, 0) for d in results) / sum(1 for d in results if k in d)
            for k in {k for d in results for k in d}
        }
        # Add group counting metrics
        data["num_groups_submitted"] = num_groups_submitted
        data["num_groups_trainable"] = num_groups_trainable
        # Get the current step after training
        current_step = self.__get_step(model)
        self._log_metrics(model, data, "train", step=current_step)
        if verbose:
            print("_train_model complete")

    def _get_reward_std_dev_learning_rate_multiplier(
        self, model: TrainableModel
    ) -> float:
        output_dir = get_model_dir(model=model, art_path=self._path)
        learning_rate_multiplier = 1.0  # Default prior
        try:
            std_dev_history = (
                pl.read_ndjson(f"{output_dir}/history.jsonl")
                .drop_nulls(subset=["train/reward_std_dev"])
                .group_by("step")
                .mean()
                .sort("step")
            )

            # Fit linear regression to std_dev_history
            if len(std_dev_history) > 1:
                steps = std_dev_history["step"].to_numpy()
                std_devs = std_dev_history["train/reward_std_dev"].to_numpy()

                # Fit linear regression: y = mx + b
                # polyfit returns [coefficient, intercept] for degree 1
                coefficient, intercept = np.polyfit(steps, std_devs, deg=1)

                # Get prediction for the last step
                last_step = steps[-1]
                last_step_prediction = coefficient * last_step + intercept
                last_step_actual = std_devs[-1]

                # Calculate R-squared and adjusted R-squared
                predictions = coefficient * steps + intercept
                ss_residual = np.sum((std_devs - predictions) ** 2)
                ss_total = np.sum((std_devs - np.mean(std_devs)) ** 2)
                r_squared = 1 - (ss_residual / ss_total) if ss_total > 0 else 0

                # Adjusted R-squared accounts for sample size
                # For simple linear regression: adj_R² = 1 - (1 - R²) * (n - 1) / (n - 2)
                n_samples = len(steps)
                if n_samples > 2:
                    adjusted_r_squared = 1 - (1 - r_squared) * (n_samples - 1) / (
                        n_samples - 2
                    )
                else:
                    adjusted_r_squared = (
                        0  # Not enough samples for meaningful adjustment
                    )

                # Calculate learning rate multiplier
                # raw_multiplier = last_step_prediction / intercept (if intercept > 0)
                # adjusted by goodness of fit: multiplier = 1 + adj_R² * (raw_multiplier - 1)
                if intercept > 0:
                    raw_multiplier = last_step_prediction / intercept
                    # learning_rate_multiplier = 1 + adjusted_r_squared * (
                    #     raw_multiplier - 1
                    # )
                    learning_rate_multiplier = raw_multiplier
                else:
                    # If intercept <= 0, can't calculate meaningful ratio, stick with prior
                    raw_multiplier = 1.0
                    learning_rate_multiplier = 1.0

                print(f"Regression fitted: y = {coefficient:.6f}x + {intercept:.6f}")
                print(f"  Coefficient (slope): {coefficient:.6f}")
                print(f"  Intercept: {intercept:.6f}")
                print(f"  R-squared: {r_squared:.4f}")
                print(
                    f"  Adjusted R-squared: {adjusted_r_squared:.4f} (n={n_samples} samples)"
                )
                print(
                    f"  Last step ({last_step}) prediction: {last_step_prediction:.6f}"
                )
                print(f"  Last step actual value: {last_step_actual:.6f}")
                print(
                    f"  Prediction error: {abs(last_step_actual - last_step_prediction):.6f}"
                )
                print(f"  Raw LR multiplier (pred/intercept): {raw_multiplier:.4f}")
                print(f"  Adjusted LR multiplier: {learning_rate_multiplier:.4f}")
            else:
                print(
                    f"Not enough data points to fit regression (need at least 2, got {len(std_dev_history)})"
                )

        except FileNotFoundError:
            print(f'"{output_dir}/history.jsonl" not found')
        except pl.exceptions.ColumnNotFoundError:
            print(f'No "train/reward_std_dev" metric found in history')

        return learning_rate_multiplier

    def _log_metrics(
        self,
        model: Model,
        metrics: dict[str, float],
        split: str,
        step: int | None = None,
    ) -> None:
        metrics = {f"{split}/{metric}": value for metric, value in metrics.items()}
        step = step if step is not None else self.__get_step(model)

        with open(
            f"{get_model_dir(model=model, art_path=self._path)}/history.jsonl", "a"
        ) as f:
            f.write(
                json.dumps(
                    {
                        k: v for k, v in metrics.items() if v == v
                    }  # Filter out NaN values
                    | {"step": step, "recorded_at": datetime.now().isoformat()}
                )
                + "\n"
            )

        # If we have a W&B run, log the data there
        if run := self._get_wandb_run(model):
            # Mark the step metric itself as hidden so W&B doesn't create an automatic chart for it
            wandb.define_metric("training_step", hidden=True)

            # Enabling the following line will cause W&B to use the training_step metric as the x-axis for all metrics
            # wandb.define_metric(f"{split}/*", step_metric="training_step")
            run.log({"training_step": step, **metrics}, step=step)

    def _get_wandb_run(self, model: Model) -> Run | None:
        if "WANDB_API_KEY" not in os.environ:
            return None
        if (
            model.name not in self._wandb_runs
            or self._wandb_runs[model.name]._is_finished
        ):
            run = wandb.init(
                project=model.project,
                name=model.name,
                id=model.name,
                resume="allow",
                settings=wandb.Settings(
                    x_stats_open_metrics_endpoints={
                        "vllm": "http://localhost:8000/metrics",
                    },
                    x_stats_open_metrics_filters=(
                        "vllm.vllm:num_requests_waiting",
                        "vllm.vllm:num_requests_running",
                    ),
                ),
            )
            self._wandb_runs[model.name] = run
            os.environ["WEAVE_PRINT_CALL_LINK"] = os.getenv(
                "WEAVE_PRINT_CALL_LINK", "False"
            )
            os.environ["WEAVE_LOG_LEVEL"] = os.getenv("WEAVE_LOG_LEVEL", "CRITICAL")
            self._weave_clients[model.name] = weave.init(model.project)
        return self._wandb_runs[model.name]

    # ------------------------------------------------------------------
    # Experimental support for S3
    # ------------------------------------------------------------------

    async def _experimental_pull_from_s3(
        self,
        model: Model,
        *,
        s3_bucket: str | None = None,
        prefix: str | None = None,
        verbose: bool = False,
        delete: bool = False,
        only_step: int | Literal["latest"] | None = None,
        # LocalBackend extensions (not part of the base interface)
        step: int | None = None,
        exclude: list[ExcludableOption] | None = None,
        latest_only: bool = False,
    ) -> None:
        """Download the model directory from S3 into local Backend storage. Right now this can be used to pull trajectory logs for processing or model checkpoints.
        Args:
            model: The model to pull from S3.
            step: DEPRECATED. Use only_step instead.
            s3_bucket: The S3 bucket to pull from. If None, the default bucket will be used.
            prefix: The prefix to pull from S3. If None, the model name will be used.
            verbose: Whether to print verbose output.
            delete: Whether to delete the local model directory.
            exclude: List of directories to exclude from sync. Valid options: "checkpoints", "logs", "trajectories".
            latest_only: DEPRECATED. Use only_step="latest" instead.
            only_step: If specified, only pull this specific step. Can be an int for a specific step,
                      or "latest" to pull only the latest checkpoint. If None, pulls all steps.
        """

        # Handle backward compatibility and new only_step parameter
        if only_step is None and latest_only:
            only_step = "latest"

        # Handle the only_step parameter
        if only_step is not None and step is None:
            if only_step == "latest":
                from art.utils.s3_checkpoint_utils import (
                    get_latest_checkpoint_step_from_s3,
                )

                latest_step = await get_latest_checkpoint_step_from_s3(
                    model_name=model.name,
                    project=model.project,
                    s3_bucket=s3_bucket,
                    prefix=prefix,
                )

                if latest_step is not None:
                    step = latest_step
                    if verbose:
                        print(f"Found latest checkpoint at step {step}")
                else:
                    if verbose:
                        print("No checkpoints found in S3")
                    return
            else:
                # only_step is an int
                step = only_step
                if verbose:
                    print(f"Pulling specific checkpoint at step {step}")

        await pull_model_from_s3(
            model_name=model.name,
            project=model.project,
            step=step,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            delete=delete,
            art_path=self._path,
            exclude=exclude,
        )

    async def _experimental_push_to_s3(
        self,
        model: Model,
        *,
        s3_bucket: str | None = None,
        prefix: str | None = None,
        verbose: bool = False,
        delete: bool = False,
    ) -> None:
        """Upload the model directory from local storage to S3."""
        await push_model_to_s3(
            model_name=model.name,
            project=model.project,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            delete=delete,
            art_path=self._path,
        )

    async def _experimental_fork_checkpoint(
        self,
        model: Model,
        from_model: str,
        from_project: str | None = None,
        from_s3_bucket: str | None = None,
        not_after_step: int | None = None,
        verbose: bool = False,
        prefix: str | None = None,
    ) -> None:
        """Fork a checkpoint from another model to initialize this model.

        Args:
            model: The model to fork to.
            from_model: The name of the model to fork from.
            from_project: The project of the model to fork from. Defaults to model.project.
            from_s3_bucket: Optional S3 bucket to pull the checkpoint from. If provided,
                will pull from S3 first. Otherwise, will fork from local disk.
            not_after_step: Optional step number. If provided, will copy the last saved
                checkpoint that is <= this step. Otherwise, copies the latest checkpoint.
            verbose: Whether to print verbose output.
            prefix: Optional S3 prefix for the bucket.
        """
        # Default from_project to model.project if not provided
        from_project = from_project or model.project

        # Get source and destination directories
        source_model_dir = get_output_dir_from_model_properties(
            project=from_project,
            name=from_model,
            art_path=self._path,
        )
        dest_model_dir = get_output_dir_from_model_properties(
            project=model.project,
            name=model.name,
            art_path=self._path,
        )

        # If S3 bucket is provided, pull from S3 first
        if from_s3_bucket is not None:
            if verbose:
                print(
                    f"DEBUG: Fork checkpoint - from_s3_bucket={from_s3_bucket}, not_after_step={not_after_step}"
                )

            # Determine which checkpoint to pull
            if not_after_step is None:
                # Pull only the latest checkpoint
                if verbose:
                    print(
                        f"Pulling latest checkpoint for model {from_model} from S3 bucket {from_s3_bucket}..."
                    )
                await self._experimental_pull_from_s3(
                    Model(name=from_model, project=from_project),
                    s3_bucket=from_s3_bucket,
                    verbose=verbose,
                    exclude=["logs", "trajectories"],
                    only_step="latest",
                )
            else:
                # Find the right checkpoint not after the specified step
                from art.utils.s3_checkpoint_utils import (
                    get_checkpoint_step_not_after_from_s3,
                )

                if verbose:
                    print(
                        f"Finding checkpoint not after step {not_after_step} for model {from_model} in S3..."
                    )

                # Find which step to pull
                target_step = await get_checkpoint_step_not_after_from_s3(
                    model_name=from_model,
                    project=from_project,
                    not_after_step=not_after_step,
                    s3_bucket=from_s3_bucket,
                    prefix=prefix,
                )

                if target_step is None:
                    raise ValueError(
                        f"No checkpoints found not after step {not_after_step} for model {from_model} in S3"
                    )

                if verbose:
                    print(
                        f"Found checkpoint at step {target_step}, pulling only this checkpoint..."
                    )

                # Pull only the specific checkpoint we need
                await pull_model_from_s3(
                    model_name=from_model,
                    project=from_project,
                    step=target_step,
                    s3_bucket=from_s3_bucket,
                    verbose=verbose,
                    art_path=self._path,
                    exclude=["logs", "trajectories"],  # Only need checkpoints
                )

        # Find the checkpoint to fork
        checkpoint_base_dir = os.path.join(source_model_dir, "checkpoints")
        if not os.path.exists(checkpoint_base_dir):
            raise FileNotFoundError(
                f"No checkpoints found for model {from_model} in project {from_project}"
            )

        if verbose:
            print(f"DEBUG: Checkpoint base dir: {checkpoint_base_dir}")
            print(
                f"DEBUG: Contents: {os.listdir(checkpoint_base_dir) if os.path.exists(checkpoint_base_dir) else 'Does not exist'}"
            )

        # Get all available checkpoint steps
        available_steps = sorted(
            int(d)
            for d in os.listdir(checkpoint_base_dir)
            if os.path.isdir(os.path.join(checkpoint_base_dir, d)) and d.isdigit()
        )

        if not available_steps:
            raise FileNotFoundError(
                f"No checkpoint directories found for model {from_model}"
            )

        # Determine which step to use
        if not_after_step is None:
            # Use the latest checkpoint
            selected_step = available_steps[-1]
        else:
            # Find the last checkpoint not after the specified step
            valid_steps = [s for s in available_steps if s <= not_after_step]
            if not valid_steps:
                raise ValueError(
                    f"No checkpoints found not after step {not_after_step}. "
                    f"Available steps: {available_steps}"
                )
            selected_step = valid_steps[-1]

        # Create destination checkpoint directory
        dest_checkpoint_dir = get_step_checkpoint_dir(dest_model_dir, selected_step)
        os.makedirs(os.path.dirname(dest_checkpoint_dir), exist_ok=True)

        # Copy the checkpoint
        source_checkpoint_dir = os.path.join(
            checkpoint_base_dir, f"{selected_step:04d}"
        )
        if verbose:
            print(
                f"Copying checkpoint from {source_checkpoint_dir} to {dest_checkpoint_dir}"
            )
            print(f"DEBUG: Source dir exists: {os.path.exists(source_checkpoint_dir)}")
            if os.path.exists(source_checkpoint_dir):
                print(
                    f"DEBUG: Source dir contents: {os.listdir(source_checkpoint_dir)}"
                )
                print(
                    f"DEBUG: Source dir is empty: {len(os.listdir(source_checkpoint_dir)) == 0}"
                )

        import shutil

        # Remove destination if it already exists (empty directory from previous attempts)
        if os.path.exists(dest_checkpoint_dir):
            if verbose:
                print("DEBUG: Destination already exists, removing it first")
            shutil.rmtree(dest_checkpoint_dir)

        shutil.copytree(source_checkpoint_dir, dest_checkpoint_dir)

        if verbose:
            print(
                f"Successfully forked checkpoint from {from_model} (step {selected_step}) to {model.name}"
            )

    async def _experimental_deploy(
        self,
        deploy_to: LoRADeploymentProvider,
        model: "TrainableModel",
        step: int | None = None,
        s3_bucket: str | None = None,
        prefix: str | None = None,
        verbose: bool = False,
        pull_s3: bool = True,
        wait_for_completion: bool = True,
    ) -> LoRADeploymentJob:
        """
        Deploy the model's latest checkpoint to a hosted inference endpoint.

        Together is currently the only supported provider. See link for supported base models:
        https://docs.together.ai/docs/lora-inference#supported-base-models
        """
        return await deploy_model(
            deploy_to=deploy_to,
            model=model,
            step=step,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            pull_s3=pull_s3,
            wait_for_completion=wait_for_completion,
            art_path=self._path,
        )



"""MCP Server wrapper class for integration with ART."""

import json
import ssl
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import aiohttp
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client


class McpServer(ABC):
    """Abstract base class for all MCP server implementations."""

    def __init__(self):
        """Initialize MCP server."""
        pass

    @abstractmethod
    async def start(self) -> None:
        """Start the MCP server connection."""
        pass

    @abstractmethod
    async def stop(self) -> None:
        """Stop the MCP server connection."""
        pass

    @abstractmethod
    async def get_tools(self) -> List[Dict[str, Any]]:
        """Get available tools from the MCP server.

        Returns:
            List of tool schemas in OpenAI format
        """
        pass

    @abstractmethod
    async def apply_tool(self, tool_name: str, arguments: Dict[str, Any]) -> str:
        """Execute a tool with given arguments.

        Args:
            tool_name: Name of the tool to call
            arguments: Arguments to pass to the tool

        Returns:
            Tool result as string
        """
        pass

    async def __aenter__(self):
        """Async context manager entry."""
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.stop()


class LocalMcpServer(McpServer):
    """MCP server that runs locally via stdio communication."""

    def __init__(self):
        """Initialize local MCP server wrapper."""
        super().__init__()
        self._session: Optional[ClientSession] = None
        self._read = None
        self._write = None
        self._stdio_context = None
        self._session_context = None

    @abstractmethod
    async def init_server(self) -> StdioServerParameters:
        """Initialize server configuration.

        Returns:
            StdioServerParameters for the specific server implementation
        """
        pass

    async def start(self) -> None:
        """Start the MCP server and establish connection."""
        server_params = await self.init_server()

        # Start stdio client
        self._stdio_context = stdio_client(server_params)
        self._read, self._write = await self._stdio_context.__aenter__()

        # Start session
        self._session_context = ClientSession(self._read, self._write)
        self._session = await self._session_context.__aenter__()

        # Initialize session
        await self._session.initialize()

    async def stop(self) -> None:
        """Stop the MCP server and clean up connections."""
        if self._session_context:
            await self._session_context.__aexit__(None, None, None)
            self._session_context = None
            self._session = None

        if self._stdio_context:
            await self._stdio_context.__aexit__(None, None, None)
            self._stdio_context = None
            self._read = None
            self._write = None

    async def get_tools(self) -> List[Dict[str, Any]]:
        """Get available tools in OpenAI format.

        Returns:
            List of tool schemas in OpenAI format
        """
        if not self._session:
            raise RuntimeError("Server not started. Call start() first.")

        # Get available tools from MCP server
        tools_result = await self._session.list_tools()

        # Convert to OpenAI format
        tool_schemas = []
        for tool in tools_result.tools:
            tool_schema = {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description or f"MCP tool: {tool.name}",
                    "parameters": tool.inputSchema
                    or {"type": "object", "properties": {}},
                },
            }
            tool_schemas.append(tool_schema)

        return tool_schemas

    async def apply_tool(self, tool_name: str, arguments: Dict[str, Any]) -> str:
        """Apply a tool with given arguments.

        Args:
            tool_name: Name of the tool to call
            arguments: Arguments to pass to the tool

        Returns:
            Tool result as string

        Raises:
            RuntimeError: If server not started
        """
        if not self._session:
            raise RuntimeError("Server not started. Call start() first.")

        result = await self._session.call_tool(tool_name, arguments)
        # Extract text content from MCP result
        if hasattr(result, "content") and result.content:
            if isinstance(result.content, list):
                # Handle list of content items
                content_text = ""
                for item in result.content:
                    if hasattr(item, "text"):
                        content_text += item.text
                    else:
                        content_text += str(item)
            else:
                content_text = (
                    result.content.text
                    if hasattr(result.content, "text")
                    else str(result.content)
                )
        else:
            content_text = str(result)

        return content_text


class RemoteMcpServer(McpServer):
    """MCP server that communicates with a hosted API server via HTTP."""

    def __init__(self, api_endpoint: str):
        """Initialize API-based MCP server.

        Args:
            api_endpoint: Base URL of the hosted MCP server API
        """
        super().__init__()
        self.api_endpoint = api_endpoint.rstrip("/")
        self._session: Optional[aiohttp.ClientSession] = None

    async def start(self) -> None:
        """Start the API session."""
        # Create SSL context that doesn't verify certificates (for development)
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        connector = aiohttp.TCPConnector(ssl=ssl_context)
        self._session = aiohttp.ClientSession(connector=connector)

    async def stop(self) -> None:
        """Stop the API session."""
        if self._session:
            await self._session.close()
            self._session = None

    async def get_tools(self) -> List[Dict[str, Any]]:
        """Get available tools from the API server.

        Returns:
            List of tool schemas in OpenAI format
        """
        if not self._session:
            raise RuntimeError("Server not started. Call start() first.")

        async with self._session.get(f"{self.api_endpoint}/tools") as response:
            response.raise_for_status()
            tools_data = await response.json()

            # Convert to OpenAI format if needed
            tool_schemas = []
            for tool in tools_data.get("tools", []):
                if "function" in tool:
                    # Already in OpenAI format
                    tool_schemas.append(tool)
                else:
                    # Convert from MCP format to OpenAI format
                    tool_schema = {
                        "type": "function",
                        "function": {
                            "name": tool.get("name", ""),
                            "description": tool.get(
                                "description", f"API tool: {tool.get('name', '')}"
                            ),
                            "parameters": tool.get(
                                "inputSchema", {"type": "object", "properties": {}}
                            ),
                        },
                    }
                    tool_schemas.append(tool_schema)

            return tool_schemas

    async def apply_tool(self, tool_name: str, arguments: Dict[str, Any]) -> str:
        """Execute a tool via the API server.

        Args:
            tool_name: Name of the tool to call
            arguments: Arguments to pass to the tool

        Returns:
            Tool result as string

        Raises:
            RuntimeError: If server not started
        """
        if not self._session:
            raise RuntimeError("Server not started. Call start() first.")

        payload = {"tool_name": tool_name, "arguments": arguments}

        async with self._session.post(
            f"{self.api_endpoint}/execute",
            json=payload,
            headers={"Content-Type": "application/json"},
        ) as response:
            response.raise_for_status()
            result_data = await response.json()

            # Extract result text
            if "result" in result_data:
                if isinstance(result_data["result"], str):
                    return result_data["result"]
                elif isinstance(result_data["result"], dict):
                    # Handle structured result
                    if "content" in result_data["result"]:
                        content = result_data["result"]["content"]
                        if isinstance(content, list):
                            # Handle list of content items
                            content_text = ""
                            for item in content:
                                if isinstance(item, dict) and "text" in item:
                                    content_text += item["text"]
                                else:
                                    content_text += str(item)
                            return content_text
                        elif isinstance(content, dict) and "text" in content:
                            return content["text"]
                        else:
                            return str(content)
                    else:
                        return json.dumps(result_data["result"])
                else:
                    return str(result_data["result"])
            else:
                return json.dumps(result_data)


class AlphaMcpServer(LocalMcpServer):
    """AlphaVantage MCP server implementation."""

    def __init__(self, api_key: str = "demo"):
        """Initialize AlphaVantage MCP server.

        Args:
            api_key: AlphaVantage API key (defaults to 'demo')
        """
        super().__init__()
        self.api_key = api_key

    async def init_server(self) -> StdioServerParameters:
        """Initialize AlphaVantage server configuration.

        Returns:
            StdioServerParameters for AlphaVantage server
        """
        return StdioServerParameters(
            command="python",
            args=[
                "servers/python/mcp_alphavantage/server.py",
                "--api-key",
                self.api_key,
            ],
            env={"ALPHAVANTAGE_API_KEY": self.api_key},
        )



import contextvars
import json
from typing import Any, AsyncIterator, Coroutine, Iterator, Literal, overload

import httpx._models
from openai import OpenAI
from openai._streaming import Stream
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from .openai import consume_sync_chat_completion_stream
from .trajectories import History, Trajectory


@overload
def auto_trajectory(*, required: Literal[True]) -> Trajectory: ...


@overload
def auto_trajectory(*, required: Literal[False] = False) -> Trajectory | None: ...


def auto_trajectory(*, required: bool = False) -> Trajectory | None:
    context = auto_trajectory_context_var.get(None)
    if context is None:
        if required:
            raise RuntimeError(
                "No auto trajectory in context. `auto_trajectory(required=True)` must be called in a `capture_auto_trajectory(...)` scope."
            )
        return None
    return context.trajectory


async def capture_auto_trajectory(coroutine: Coroutine[Any, Any, Any]) -> Trajectory:
    with AutoTrajectoryContext():
        await coroutine
        trajectory = auto_trajectory_context_var.get().trajectory
        trajectory.finish()
        return trajectory


class AutoTrajectoryContext:
    def __init__(self) -> None:
        self.trajectory = Trajectory(
            messages_and_choices=[],
            reward=0.0,
        )
        self.openai_client = OpenAI(api_key="")

    def __enter__(self) -> None:
        self.token = auto_trajectory_context_var.set(self)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        auto_trajectory_context_var.reset(self.token)

    def handle_httpx_response(self, response: httpx._models.Response) -> None:
        try:
            request_content = json.loads(getattr(response.request, "_content", b""))
            messages = request_content["messages"]
            tools = request_content.get("tools", None)
            setattr(response, "_content", getattr(response, "_content_so_far", b""))
            print(getattr(response, "_content"))
            if request_content.get("stream", False):
                choice = consume_sync_chat_completion_stream(
                    Stream(
                        cast_to=ChatCompletionChunk,
                        response=response,
                        client=self.openai_client,
                    )
                ).choices[0]
            else:
                choice = Choice(
                    **json.loads(getattr(response, "_content"))["choices"][0]
                )
            history: Trajectory | History = self.trajectory
            history_index = -1
            while True:
                history_messages = history.messages()
                if history_messages == messages[: len(history_messages)] and (
                    history.tools == tools
                    or (history_messages == [] and history.tools is None)
                ):
                    break
                history_index += 1
                try:
                    history = self.trajectory.additional_histories[history_index]
                except IndexError:
                    history = History(messages_and_choices=[])
                    self.trajectory.additional_histories.append(history)
                    break
            history.messages_and_choices.extend(
                messages[len(history.messages_and_choices) :]
            )
            history.messages_and_choices.append(choice)
            history.tools = tools
        except:
            pass


auto_trajectory_context_var: contextvars.ContextVar[AutoTrajectoryContext] = (
    contextvars.ContextVar("auto_trajectory_context")
)


def patch_httpx() -> None:
    original_iter_bytes = httpx._models.Response.iter_bytes
    original_aiter_bytes = httpx._models.Response.aiter_bytes
    original_close = httpx._models.Response.close
    original_aclose = httpx._models.Response.aclose

    def patched_iter_bytes(
        self: httpx._models.Response, chunk_size: int | None = None
    ) -> Iterator[bytes]:
        for chunk in original_iter_bytes(self, chunk_size):
            setattr(
                self, "_content_so_far", getattr(self, "_content_so_far", b"") + chunk
            )
            yield chunk

    async def patched_aiter_bytes(
        self: httpx._models.Response, chunk_size: int | None = None
    ) -> AsyncIterator[bytes]:
        async for chunk in original_aiter_bytes(self, chunk_size):
            setattr(
                self, "_content_so_far", getattr(self, "_content_so_far", b"") + chunk
            )
            yield chunk

    def patched_close(self: httpx._models.Response) -> None:
        original_close(self)
        if context := auto_trajectory_context_var.get(None):
            context.handle_httpx_response(self)

    async def patched_aclose(self: httpx._models.Response) -> None:
        await original_aclose(self)
        if context := auto_trajectory_context_var.get(None):
            context.handle_httpx_response(self)

    httpx._models.Response.iter_bytes = patched_iter_bytes
    httpx._models.Response.aiter_bytes = patched_aiter_bytes
    httpx._models.Response.close = patched_close
    httpx._models.Response.aclose = patched_aclose


patch_httpx()



import asyncio
import glob
import logging
import os
import time
from collections import Counter
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import AsyncIterator

import torch
import torchtune
from safetensors.torch import load_file
from vllm import AsyncEngineArgs
from vllm.v1.engine.async_llm import AsyncLLM

from .. import dev, types
from ..preprocessing.pack import DiskPackedTensors
from ..vllm import get_llm, get_worker, openai_server_task, run_on_workers
from .batch import Batch


@dataclass
class TorchtuneService:
    model_name: str
    base_model: str
    config: dev.InternalModelConfig
    output_dir: str
    _is_sleeping: bool = False

    async def start_openai_server(self, config: dev.OpenAIServerConfig | None) -> None:
        await openai_server_task(
            engine=await self.llm,
            config=dev.get_openai_server_config(
                model_name=self.model_name,
                base_model=self.get_last_checkpoint_dir() or self.base_model,
                log_file=f"{self.output_dir}/logs/vllm.log",
                config=config,
            ),
        )

    async def vllm_engine_is_sleeping(self) -> bool:
        return self._is_sleeping

    async def train(
        self,
        disk_packed_tensors: DiskPackedTensors,
        config: types.TrainConfig,
        _config: dev.TrainConfig,
        verbose: bool = False,
    ) -> AsyncIterator[dict[str, float]]:
        llm = await self.llm
        pids_path = f"{self.output_dir}/pids.txt"
        # reset the pids file
        with open(pids_path, "w") as f:
            f.write("")
        weights_path = "/dev/shm/weights.safetensors"
        # remove the weights file if it exists
        Path(weights_path).unlink(missing_ok=True)
        async_weight_syncing = self.torchtune_args.get("async_weight_syncing", False)
        # start putting the workers to sleep
        sleep_task = asyncio.create_task(
            run_on_workers(
                llm,
                sleep,
                # level=1 if llm.output_processor.has_unfinished_requests() else 2,
                level=1,
                pids_path=pids_path,
                weights_path=None if async_weight_syncing else weights_path,
                profile=verbose,
            )
        )
        # wait for the workers to write their pids twice, indicating that they are asleep
        while True:
            pids = Counter(open(pids_path).read().splitlines())
            if set(pids.values()) == {2}:
                break
            await asyncio.sleep(0.25)
        self._is_sleeping = True
        # acquire the train process and queue
        train_process = await self.train_process
        train_queue = await self.train_queue
        # write the batch to communicate with the train process
        with open(f"{self.output_dir}/batches.jsonl", "a") as f:
            f.write(
                Batch(
                    disk_packed_tensors=disk_packed_tensors,
                    config=config,
                    dev_config=_config,
                ).model_dump_json()
                + "\n"
            )
        # consume the batch gradient step results
        num_gradient_steps = -1
        while num_gradient_steps != 0:
            done, _ = await asyncio.wait(
                [
                    asyncio.create_task(train_queue.get()),
                    asyncio.create_task(train_process.wait()),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in done:
                result = task.result()
                if isinstance(result, dict):
                    result["num_gradient_steps"] = int(result["num_gradient_steps"])
                    if num_gradient_steps == -1:
                        num_gradient_steps = result["num_gradient_steps"]
                    yield result
                else:
                    raise RuntimeError(
                        f"Train process exited early. See {self.output_dir}/logs/train.log for details."
                    )
            num_gradient_steps -= 1
        # wait for the workers to wake up
        await sleep_task
        self._is_sleeping = False
        # update the weights after wake up if async_weight_syncing is enabled
        if async_weight_syncing:
            asyncio.create_task(self.update_worker_weights(llm, weights_path, verbose))
        else:
            # remove the weights file
            Path(weights_path).unlink(missing_ok=True)

    async def update_worker_weights(
        self, llm: AsyncLLM, weights_path: str, profile: bool
    ) -> None:
        while True:
            if os.path.exists(weights_path):
                break
            else:
                time.sleep(1)
                continue
        await run_on_workers(
            llm,
            update_weights,
            weights_path=weights_path,
            profile=profile,
        )
        # remove the weights file
        Path(weights_path).unlink(missing_ok=True)

    @property
    def torchtune_args(self) -> dev.TorchtuneArgs:
        torchtune_args = self.config.get("torchtune_args")
        assert torchtune_args is not None, (
            'TorchtuneService created without config["torchtune_args"]'
        )
        return torchtune_args

    @cached_property
    def llm(self) -> asyncio.Task[AsyncLLM]:
        return asyncio.create_task(
            get_llm(AsyncEngineArgs(**self.config.get("engine_args", {})))  # type: ignore
        )

    @cached_property
    def train_queue(self) -> asyncio.Task[asyncio.Queue[dict[str, float]]]:
        return asyncio.create_task(self.get_train_queue())

    @cached_property
    def train_process(self) -> asyncio.Task[asyncio.subprocess.Process]:
        return asyncio.create_task(self.get_train_process())

    async def get_train_process(self) -> asyncio.subprocess.Process:
        # Migrate existing checkpoints to new structure if needed
        from ..local.checkpoints import migrate_checkpoints_to_new_structure

        migrate_checkpoints_to_new_structure(self.output_dir)

        Path(f"{self.output_dir}/batches.jsonl").unlink(missing_ok=True)
        checkpoint_dir = await self.get_checkpoint_dir()
        torchtune_args = self.torchtune_args

        # Get the list of safetensor files
        safetensor_files = glob.glob(f"{checkpoint_dir}/*.safetensors")
        checkpoint_files = [os.path.basename(f) for f in safetensor_files]
        checkpoint_files_str = "[" + ", ".join(f'"{f}"' for f in checkpoint_files) + "]"

        def model_dir(model: str) -> str:
            for prefix in [
                "llama3_1",
                "llama3_2_vision",
                "llama3_2",
                "llama3_3",
                "qwen2_5",
            ]:
                if model.startswith(prefix):
                    return prefix
            return model.split("_")[0]

        program_and_args = [
            "python",  # Use Python interpreter
            f"{os.path.dirname(torchtune.__file__)}/_cli/tune.py",
            "run",
            "--nproc-per-node",
            str(torch.cuda.device_count()),
            "art.torchtune.recipe.FullFinetuneRecipeDistributed",
            "--config",
            f"{os.path.dirname(__file__)}/config.yaml",
            f"model._component_=torchtune.models.{model_dir(torchtune_args['model'])}.{torchtune_args['model']}",
            f"checkpointer.checkpoint_dir={checkpoint_dir}",
            f"checkpointer.checkpoint_files={checkpoint_files_str}",
            f"checkpointer.model_type={torchtune_args['model_type']}",
            f"tensor_parallel_dim={torchtune_args.get('tensor_parallel_dim', 1)}",
            f"context_parallel_dim={torchtune_args.get('context_parallel_dim', 1)}",
            f"output_dir={self.output_dir}",
            "clip_grad_norm=0.1",
            "metric_logger._component_=torchtune.training.metric_logging.StdoutLogger",
            "metric_logger.log_dir=null",
            f"enable_activation_offloading={torchtune_args.get('enable_activation_offloading', False)}",
        ]
        return await asyncio.subprocess.create_subprocess_exec(
            *program_and_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

    async def get_train_queue(self) -> asyncio.Queue[dict[str, float]]:
        process = await self.train_process
        queue = asyncio.Queue()

        async def read(reader: asyncio.StreamReader) -> None:
            async for line in reader:
                line_str = line.decode("utf-8")
                with open(f"{self.output_dir}/logs/train.log", "a") as f:
                    f.write(line_str)
                line_str = line_str.strip()
                if line_str.startswith("Step ") and " | " in line_str:
                    parts = line_str.split(" | ", 1)
                    metrics: dict[str, float] = {}
                    if len(parts) > 1:
                        for metric in parts[1].split():
                            if ":" in metric:
                                name, value = metric.split(":", 1)
                                try:
                                    metrics[name] = float(value)
                                except ValueError:
                                    # Skip non-numeric values to match the return type
                                    pass
                    await queue.put(metrics)

        assert process.stdout and process.stderr
        asyncio.create_task(read(process.stdout))
        asyncio.create_task(read(process.stderr))
        return queue

    async def get_checkpoint_dir(self) -> str:
        # Use the last of any existing checkpoints to resume training
        if last_checkpoint_dir := self.get_last_checkpoint_dir():
            return last_checkpoint_dir
        # Check if self.base_model is a directory
        if os.path.isdir(self.base_model):
            return self.base_model
        # Otherwise, assume it's a HuggingFace model id and download it
        process = await asyncio.subprocess.create_subprocess_exec(
            "huggingface-cli",
            "download",
            self.base_model,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, _ = await process.communicate()
        return stdout.decode("utf-8").splitlines()[-1].strip()

    def get_last_checkpoint_dir(self) -> str | None:
        from ..local.checkpoints import get_last_checkpoint_dir

        return get_last_checkpoint_dir(self.output_dir)


def sleep(
    *, level: int, pids_path: str, weights_path: str | None, profile: bool
) -> None:
    """
    Put the worker to sleep until the new model weights are loaded.

    Args:
        level: The sleep level: 1 to offload the kv cache, 2 to discard the kv cache.
        pids_path: The path to the file that contains the PIDs of the workers.
        weights_path: The path to the weights file.
        profile: Whether to profile
    """
    from vllm.device_allocator.cumem import CuMemAllocator
    from vllm.v1.worker.gpu_worker import logger

    with open(pids_path, "a") as f:
        f.write(f"{os.getpid()}\n")
    worker = get_worker()
    allocator = CuMemAllocator.get_instance()
    try:
        if not (profile and worker.rank == 0):
            logger.setLevel(logging.CRITICAL)
        setattr(allocator, "_override_tags", {"weights", "kv_cache"})
        with worker.time("sleep"):
            worker.sleep(level)
        with open(pids_path, "a") as f:
            f.write(f"{os.getpid()}\n")
        weights = None
        while True:
            if weights_path:
                # wait for the weights file to be created
                try:
                    with worker.time("load_file"):
                        weights = load_file(weights_path)
                    break
                except FileNotFoundError:
                    time.sleep(1)
                    continue
            elif os.path.exists(pids_path):
                time.sleep(1)
                continue
            else:
                # no pids file indicates we can wake up
                break
        with worker.time("wake_up"):
            worker.wake_up()
        if weights is None:
            return
        with worker.time("load_weights"):
            worker.model_runner.model.load_weights(weights.items())  # type: ignore
    finally:
        logger.setLevel(logging.INFO)
        delattr(allocator, "_override_tags")


def update_weights(weights_path: str, profile: bool) -> None:
    from vllm.v1.worker.gpu_worker import logger

    worker = get_worker()
    try:
        if not (profile and worker.rank == 0):
            logger.setLevel(logging.CRITICAL)
        with worker.time("load_file"):
            weights = load_file(weights_path)
        with worker.time("load_weights"):
            worker.model_runner.model.load_weights(weights.items())  # type: ignore
    finally:
        logger.setLevel(logging.INFO)



import argparse
import asyncio
import os

from dotenv import load_dotenv

import art
from art.utils.deploy_model import deploy_model
from art.utils.get_model_step import get_model_step
from art.utils.s3 import pull_model_from_s3

load_dotenv()

# ---------------------------------------------------------------------------
# Argument parsing
# ---------------------------------------------------------------------------


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Deploy a model checkpoint using ART & SkyPilot"
    )

    parser.add_argument("--project", required=True, help="ART project name")
    parser.add_argument("--model", required=True, help="Name of the model to deploy")
    parser.add_argument("--base-model", required=True, help="Base model to use")

    # Optional arguments
    parser.add_argument(
        "--backup-bucket",
        help="Name of the S3 bucket containing model checkpoints",
    )
    parser.add_argument(
        "--step",
        type=str,
        default="latest",
        help="Training step to deploy (should correspond to a saved checkpoint)",
    )
    parser.add_argument(
        "--art-path", type=str, help="Path to the ART directory", default=".art"
    )

    return parser.parse_args()


# ---------------------------------------------------------------------------
# Main deployment routine
# ---------------------------------------------------------------------------


async def deploy() -> None:
    args = parse_args()

    backup_bucket = args.backup_bucket or os.environ["BACKUP_BUCKET"]

    model = art.TrainableModel(
        name=args.model,
        project=args.project,
        base_model=args.base_model,
    )

    if args.step == "latest":
        print("Pulling all checkpoints to determine the latest step…")
        # pull all checkpoints to determine the latest step
        await pull_model_from_s3(
            model_name=model.name,
            project=model.project,
            art_path=args.art_path,
            s3_bucket=backup_bucket,
        )
        step = get_model_step(model, args.art_path)
    else:
        print(f"Pulling checkpoint for step {args.step}…")
        step = int(args.step)
        # only pull the checkpoint we need
        await pull_model_from_s3(
            model_name=model.name,
            project=model.project,
            art_path=args.art_path,
            s3_bucket=backup_bucket,
            step=step,
        )

    print(
        f"Deploying {args.model} (project={args.project}, step={step}) "
        f"using checkpoints from s3://{backup_bucket}…"
    )

    deployment_result = await deploy_model(
        deploy_to="together",
        model=model,
        step=step,
        verbose=True,
        pull_s3=False,
        wait_for_completion=True,
        art_path=args.art_path,
    )

    if deployment_result.status == "Failed":
        raise RuntimeError(f"Deployment failed: {deployment_result.failure_reason}")

    print("Deployment successful! ✨")
    print(
        f"Model deployed at Together under name: {deployment_result.model_name} (job_id={deployment_result.job_id})"
    )


if __name__ == "__main__":
    asyncio.run(deploy())



#!/usr/bin/env python3

import asyncio

from just_the_facts.scenarios import train_urls, val_urls
from just_the_facts.utils import scrape_article


async def test_scraper():
    """Test the scrape_article function with example URLs"""

    # Test URLs from different news sources (using homepage URLs that should exist)
    test_urls = train_urls + val_urls

    for url in test_urls:
        try:
            print(f"\nTesting URL: {url}")
            article_text = await scrape_article(url)
            print(f"Successfully scraped {len(article_text)} characters")
            print(f"First 200 characters: {article_text[:200]}...")
        except Exception as e:
            print(f"Failed to scrape {url}: {str(e)}")
            raise e


if __name__ == "__main__":
    asyncio.run(test_scraper())



from typing import Literal

from typing_extensions import Required, TypedDict

# Update from torchtune.models.{model_family}.__init__.py files
# Only include argument-less TransformerDecoder factory functions
TorchtuneModel = Literal[
    # Gemma
    "gemma_2b",
    "gemma_7b",
    # Gemma2
    "gemma2_2b",
    "gemma2_9b",
    "gemma2_27b",
    # Llama2
    "llama2_7b",
    "llama2_13b",
    "llama2_70b",
    "llama2_reward_7b",
    # Llama3
    "llama3_8b",
    "llama3_70b",
    # Llama3.1
    "llama3_1_8b",
    "llama3_1_70b",
    "llama3_1_405b",
    # Llama3.2
    "llama3_2_1b",
    "llama3_2_3b",
    # Llama3.2 Vision
    "llama3_2_vision_11b",
    "llama3_2_vision_90b",
    # Llama3.3
    "llama3_3_70b",
    # Llama4
    "llama4_scout_17b_16e",
    "llama4_maverick_17b_128e",
    # Mistral
    "mistral_7b",
    "mistral_reward_7b",
    # Phi3
    "phi3_mini",
    # Phi4
    "phi4_14b",
    # Qwen2
    "qwen2_0_5b",
    "qwen2_1_5b",
    "qwen2_7b",
    # Qwen2.5
    "qwen2_5_0_5b",
    "qwen2_5_1_5b_base",
    "qwen2_5_1_5b_instruct",
    "qwen2_5_3b",
    "qwen2_5_7b_base",
    "qwen2_5_7b_instruct",
    "qwen2_5_14b_base",
    "qwen2_5_14b_instruct",
    "qwen2_5_32b_base",
    "qwen2_5_32b_instruct",
    "qwen2_5_72b_base",
    "qwen2_5_72b_instruct",
    # Qwen3
    "qwen3_0_6b_base",
    "qwen3_0_6b_instruct",
    "qwen3_1_7b_base",
    "qwen3_1_7b_instruct",
    "qwen3_4b_base",
    "qwen3_4b_instruct",
    "qwen3_8b_base",
    "qwen3_8b_instruct",
    "qwen3_14b_base",
    "qwen3_14b_instruct",
    "qwen3_32b",
]

# Update from torchtune.training.checkpointing._utils.ModelType
TorchtuneModelType = Literal[
    "GEMMA",
    "GEMMA2",
    "LLAMA2",
    "LLAMA3",
    "LLAMA3_2",
    "LLAMA3_VISION",
    "LLAMA4",
    "MISTRAL",
    "PHI3_MINI",
    "PHI4",
    "REWARD",
    "QWEN2",
    "CLIP_TEXT",
    "T5_ENCODER",
    "QWEN3",
]


class TorchtuneArgs(TypedDict, total=False):
    model: Required[TorchtuneModel]
    model_type: Required[TorchtuneModelType]
    tensor_parallel_dim: int
    context_parallel_dim: int
    enable_activation_offloading: bool
    async_weight_syncing: bool



import contextvars
from typing import Any, Coroutine

from .trajectories import Trajectory


def yield_trajectory(trajectory: Trajectory) -> None:
    yield_trajectory_context_var.get().trajectory = trajectory


async def capture_yielded_trajectory(coroutine: Coroutine[Any, Any, Any]) -> Trajectory:
    with YieldTrajectoryContext():
        await coroutine
        trajectory = yield_trajectory_context_var.get().trajectory
        if trajectory is None:
            raise RuntimeError("No trajectory yielded")
        return trajectory


class YieldTrajectoryContext:
    def __init__(self) -> None:
        self.trajectory: Trajectory | None = None

    def __enter__(self) -> None:
        self.token = yield_trajectory_context_var.set(self)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        yield_trajectory_context_var.reset(self.token)


yield_trajectory_context_var: contextvars.ContextVar[YieldTrajectoryContext] = (
    contextvars.ContextVar("yield_trajectory_context", default=YieldTrajectoryContext())
)



import asyncio
import logging
import os
import time
from collections import Counter
from dataclasses import dataclass
from functools import cached_property
from typing import Any, AsyncIterator, cast

import peft
import torch
from datasets import Dataset
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.utils.dummy_pt_objects import GenerationMixin, PreTrainedModel
from trl import GRPOConfig, GRPOTrainer
from vllm import AsyncEngineArgs
from vllm.device_allocator.cumem import CuMemAllocator
from vllm.lora.request import LoRARequest
from vllm.v1.engine.async_llm import AsyncLLM
from vllm.v1.worker.gpu_worker import logger

from .. import dev, types
from ..local.checkpoints import get_last_checkpoint_dir
from ..preprocessing.pack import (
    DiskPackedTensors,
    PackedTensors,
    packed_tensors_from_dir,
)
from ..utils.get_model_step import get_step_from_dir
from ..utils.output_dirs import get_step_checkpoint_dir
from ..vllm import get_llm, get_worker, openai_server_task, run_on_workers
from .train import gc_and_empty_cuda_cache, train


class CausalLM(PreTrainedModel, GenerationMixin):
    """Dummy class for type checking."""

    pass


class TrainInputs(PackedTensors):
    config: types.TrainConfig
    _config: dev.TrainConfig
    return_new_logprobs: bool


@dataclass
class UnslothState:
    model: CausalLM
    tokenizer: PreTrainedTokenizerBase
    peft_model: peft.peft_model.PeftModelForCausalLM
    trainer: GRPOTrainer
    inputs_queue: asyncio.Queue[TrainInputs]
    results_queue: asyncio.Queue[dict[str, float]]


@dataclass
class DecoupledUnslothService:
    model_name: str
    base_model: str
    config: dev.InternalModelConfig
    output_dir: str
    _is_sleeping: bool = False

    async def start_openai_server(self, config: dev.OpenAIServerConfig | None) -> None:
        lora_path = get_last_checkpoint_dir(self.output_dir)
        if lora_path is None:
            # Create initial LoRA checkpoint if none exists
            lora_path = get_step_checkpoint_dir(self.output_dir, 0)
            os.makedirs(os.path.dirname(lora_path), exist_ok=True)
            self._state.trainer.save_model(lora_path)
        await openai_server_task(
            engine=await self.llm,
            config=dev.get_openai_server_config(
                model_name=self.model_name,
                base_model=self.base_model,
                log_file=f"{self.output_dir}/logs/vllm.log",
                lora_path=lora_path,
                config=config,
            ),
        )

    async def vllm_engine_is_sleeping(self) -> bool:
        return self._is_sleeping

    async def train(
        self,
        disk_packed_tensors: DiskPackedTensors,
        config: types.TrainConfig,
        _config: dev.TrainConfig,
        verbose: bool = False,
    ) -> AsyncIterator[dict[str, float]]:
        llm = await self.llm
        pids_path = f"{self.output_dir}/pids.txt"
        # reset the pids file
        with open(pids_path, "w") as f:
            f.write("")
        # start putting the workers to sleep
        sleep_task = asyncio.create_task(
            run_on_workers(
                llm,
                sleep,
                level=1,
                pids_path=pids_path,
                profile=verbose,
            )
        )
        # wait for the workers to write their pids twice, indicating that they are asleep
        while True:
            pids = Counter(open(pids_path).read().splitlines())
            if set(pids.values()) == {2}:
                break
            await asyncio.sleep(0.25)

        self._is_sleeping = True

        # Free memory after vLLM workers are asleep
        gc_and_empty_cuda_cache()

        # Load packed tensors
        packed_tensors = packed_tensors_from_dir(**disk_packed_tensors)

        # Wait for existing batches to finish
        await self._state.results_queue.join()

        # If we haven't already, start the training task
        if not hasattr(self, "_train_task") or self._train_task is None:
            self._train_task = asyncio.create_task(
                train(
                    trainer=self._state.trainer,
                    results_queue=self._state.results_queue,
                )
            )
            warmup = True
        else:
            warmup = False
        precalculate_logprobs = _config.get("precalculate_logprobs", False)
        # Train on the batch
        for offset in range(0, packed_tensors["tokens"].shape[0]):
            for _ in range(2 if warmup else 1):
                if precalculate_logprobs and not warmup:
                    packed_tensors["logprobs"] = torch.cat(
                        [
                            self._state.trainer.compute_loss(
                                self._state.peft_model,
                                TrainInputs(
                                    **{
                                        k: v[_offset : _offset + 1]
                                        for k, v in packed_tensors.items()
                                        if isinstance(v, torch.Tensor)
                                    },
                                    config=config,
                                    _config=_config,
                                    return_new_logprobs=True,
                                ),  # type: ignore
                            )
                            for _offset in range(0, packed_tensors["tokens"].shape[0])
                        ]
                    ).to("cpu")
                    precalculate_logprobs = False
                self._state.inputs_queue.put_nowait(
                    TrainInputs(
                        **{
                            k: (
                                v[offset : offset + 1, :1024]
                                if warmup and v.dim() > 1
                                else v[offset : offset + 1]
                            )
                            for k, v in packed_tensors.items()
                            if isinstance(v, torch.Tensor)
                        },
                        config=(
                            config.model_copy(
                                update={"lr": 1e-9, "beta": 0.0, "kl_coef": 0.0}
                            )
                            if warmup
                            else config
                        ),
                        _config=_config,
                        return_new_logprobs=False,
                    )
                )
                # Wait for a result from the queue or for the training task to,
                # presumably, raise an exception
                done, _ = await asyncio.wait(
                    [
                        asyncio.create_task(self._state.results_queue.get()),
                        self._train_task,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if verbose:
                    print(
                        "Done waiting for a result from the queue or for the training task to, presumably, raise an exception"
                    )
                for task in done:
                    result = task.result()
                    # If `result` is `None`, the training task finished somehow.
                    assert result is not None, "The training task should never finish."
                    self._state.results_queue.task_done()
                    if warmup:
                        gc_and_empty_cuda_cache()
                        await asyncio.sleep(0.1)
                        warmup = False
                    else:
                        yield result

        if verbose:
            print("Saving new LoRA adapter...")
        # Save checkpoint after training
        next_step = get_step_from_dir(self.output_dir) + 1
        checkpoint_dir = get_step_checkpoint_dir(self.output_dir, next_step)
        os.makedirs(checkpoint_dir, exist_ok=True)
        self._state.trainer.save_model(checkpoint_dir)

        # Free memory before waking up vLLM
        gc_and_empty_cuda_cache()

        # Remove pids.txt to signal workers to wake up
        if os.path.exists(pids_path):
            os.remove(pids_path)
            if verbose:
                print("Removed pids.txt to signal workers to wake up")

        # wait for the workers to wake up
        await sleep_task

        self._is_sleeping = False

        # swap out the LoRA adapter
        await llm.remove_lora(1)
        await llm.add_lora(
            LoRARequest(
                lora_name=self.model_name,
                lora_int_id=1,
                lora_path=checkpoint_dir,
            )
        )

        if verbose:
            print("DecoupledUnslothService.train complete")

    @cached_property
    def _state(self) -> UnslothState:
        import unsloth

        # Initialize Unsloth model
        init_args = self.config.get("init_args", {})
        checkpoint_dir = get_last_checkpoint_dir(self.output_dir)
        if checkpoint_dir:
            init_args["model_name"] = checkpoint_dir
        else:
            init_args["model_name"] = self.base_model

        model, tokenizer = cast(
            tuple[CausalLM, PreTrainedTokenizerBase],
            unsloth.FastLanguageModel.from_pretrained(**init_args),
        )

        # Initialize PEFT model
        peft_model = cast(
            peft.peft_model.PeftModelForCausalLM,
            unsloth.FastLanguageModel.get_peft_model(
                model, **self.config.get("peft_args", {})
            ),
        )

        # Initialize trainer with dummy dataset
        data = {"prompt": ""}
        trainer = GRPOTrainer(
            model=peft_model,  # type: ignore
            reward_funcs=[],
            args=GRPOConfig(**self.config.get("trainer_args", {})),  # type: ignore
            train_dataset=Dataset.from_list([data for _ in range(10_000_000)]),
            processing_class=tokenizer,
        )

        # Initialize queues
        inputs_queue: asyncio.Queue[TrainInputs] = asyncio.Queue()
        results_queue: asyncio.Queue[dict[str, float]] = asyncio.Queue()

        # Patch trainer _prepare_inputs() to pull from queue
        def _async_prepare_inputs(*_: Any, **__: Any) -> dict[str, torch.Tensor]:
            async def get_inputs() -> TrainInputs:
                return await inputs_queue.get()

            # Force otherwise synchronous _prepare_inputs() to yield
            # with nested asyncio.run() call
            inputs = asyncio.run(get_inputs())

            return cast(dict[str, torch.Tensor], inputs)

        trainer._prepare_inputs = _async_prepare_inputs

        return UnslothState(
            model=model,
            tokenizer=tokenizer,
            peft_model=peft_model,
            trainer=trainer,
            inputs_queue=inputs_queue,
            results_queue=results_queue,
        )

    @cached_property
    def llm(self) -> asyncio.Task[AsyncLLM]:
        return asyncio.create_task(
            get_llm(
                AsyncEngineArgs(
                    **{**self.config.get("engine_args", {}), "enable_lora": True}
                )
            )
        )


def sleep(*, level: int, pids_path: str, profile: bool) -> None:
    """
    Put the worker to sleep until signaled to wake up.

    Args:
        level: The sleep level: 1 to offload the kv cache, 2 to discard the kv cache.
        pids_path: The path to the file that contains the PIDs of the workers.
        profile: Whether to profile
    """
    with open(pids_path, "a") as f:
        f.write(f"{os.getpid()}\n")
    worker = get_worker()
    allocator = CuMemAllocator.get_instance()
    try:
        if not (profile and worker.rank == 0):
            logger.setLevel(logging.CRITICAL)
        setattr(allocator, "_override_tags", {"weights", "kv_cache"})
        with worker.time("sleep"):
            worker.sleep(level)
        with open(pids_path, "a") as f:
            f.write(f"{os.getpid()}\n")

        # Wait for the signal to wake up
        while os.path.exists(pids_path):
            time.sleep(1)

        with worker.time("wake_up"):
            worker.wake_up()
    finally:
        logger.setLevel(logging.INFO)
        delattr(allocator, "_override_tags")



from typing import Literal

from typing_extensions import TypedDict


class TrainConfig(TypedDict, total=False):
    advantage_balance: float
    """Balance between negative and positive advantages in the range [-1.0, 1.0]. \
-1.0 means only training on negative advantages, 1.0 means only training on \
positive advantages. Defaults to 0.0 (perfectly balanced)."""
    allow_training_without_logprobs: bool
    epsilon: float  # clip epsilon, using the same name as TRL
    epsilon_high: (
        float | None
    )  # asymmetric clip upper bound. Defaults to epsilon when None
    importance_sampling_level: Literal["token", "sequence"]
    logprob_calculation_chunk_size: int
    max_negative_advantage_importance_sampling_weight: float
    num_trajectories_learning_rate_multiplier_power: float
    plot_tensors: bool
    precalculate_logprobs: bool
    scale_learning_rate_by_reward_std_dev: bool
    scale_rewards: bool
    truncated_importance_sampling: float | None



import json
from typing import Any, cast

import yaml

from art import Trajectory, TrajectoryGroup
from art.trajectories import History
from art.types import Choice, Message, MessageOrChoice


# serialize trajectory groups to a jsonl string
def serialize_trajectory_groups(trajectory_groups: list[TrajectoryGroup]) -> str:
    group_dicts = [
        trajectory_group_to_dict(trajectory_group)
        for trajectory_group in trajectory_groups
    ]

    return "\n".join(json.dumps(group_dict) for group_dict in group_dicts)


def trajectory_group_to_dict(trajectory_group: TrajectoryGroup) -> dict[str, Any]:
    trajectory_dicts = []
    for trajectory in trajectory_group.trajectories:
        if not isinstance(trajectory, Trajectory):
            # remove exceptions
            continue
        trajectory_dicts.append(trajectory_to_dict(trajectory))

    return {
        "trajectories": trajectory_dicts,
    }


def history_to_dict(history: History) -> dict[str, Any]:
    messages_and_choices = [
        message_or_choice_to_dict(message_or_choice)
        for message_or_choice in history.messages_and_choices
    ]

    return {"messages_and_choices": messages_and_choices, "tools": history.tools}


def trajectory_to_dict(trajectory: Trajectory) -> dict[str, Any]:
    messages_and_choices = [
        message_or_choice_to_dict(message_or_choice)
        for message_or_choice in trajectory.messages_and_choices
    ]

    return {
        "reward": trajectory.reward,
        "metrics": trajectory.metrics,
        "metadata": trajectory.metadata,
        "messages_and_choices": messages_and_choices,
        "tools": trajectory.tools,
        "additional_histories": (
            [history_to_dict(h) for h in trajectory.additional_histories]
            if trajectory.additional_histories
            else trajectory.additional_histories
        ),
        "logs": trajectory.logs,
    }


def message_or_choice_to_dict(message_or_choice: MessageOrChoice) -> dict[str, Any]:
    # messages are sometimes stored as dicts, so we need to handle both cases
    item_dict = (
        message_or_choice
        if isinstance(message_or_choice, dict)
        else message_or_choice.to_dict()
    )

    if "logprobs" in item_dict:
        # item is a choice with logprobs, remove the logprobs
        item_dict.pop("logprobs")

    return dict(item_dict)


def deserialize_trajectory_groups(serialized: str) -> list[TrajectoryGroup]:
    # Try to parse as JSONL first (new format)
    try:
        loaded_groups = [
            json.loads(line) for line in serialized.strip().split("\n") if line
        ]
    except json.JSONDecodeError:
        # Fall back to YAML parsing (old format)
        loaded_groups = yaml.load(serialized, Loader=yaml.SafeLoader)
    return [dict_to_trajectory_group(group) for group in loaded_groups]


def dict_to_trajectory_group(dict: dict[str, Any]) -> TrajectoryGroup:
    return TrajectoryGroup(
        trajectories=[
            dict_to_trajectory(trajectory) for trajectory in dict["trajectories"]
        ],
        exceptions=[],
    )


def dict_to_trajectory(dict: dict[str, Any]) -> Trajectory:
    return Trajectory(
        messages_and_choices=[
            dict_to_message_or_choice(message_or_choice)
            for message_or_choice in dict["messages_and_choices"]
        ],
        reward=dict["reward"],
        metrics=dict["metrics"],
        metadata=dict["metadata"],
        logs=dict["logs"],
    )


def dict_to_message_or_choice(dict: dict[str, Any]) -> MessageOrChoice:
    if "message" in dict:
        return Choice(**dict)
    else:
        return cast(Message, dict)



from pydantic import BaseModel

import art


class JustTheFactsConfig(BaseModel):
    learning_rate: float = 1e-6
    num_epochs: int = 20

    eval_steps: int = 1
    groups_per_step: int = 3
    trajectories_per_group: int = 4
    scale_rewards: bool = False


models: dict[str, art.TrainableModel[JustTheFactsConfig]] = {
    "facts-14b-001": art.TrainableModel(
        name="facts-14b-001",
        project="just-the-facts",
        base_model="Qwen/Qwen2.5-14B-Instruct",
        config=JustTheFactsConfig(
            num_epochs=20,
        ),
    )
}

models["facts-14b-002"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-14b-002"].name = "facts-14b-002"
models["facts-14b-002"].base_model = "Qwen/Qwen2.5-14B-Instruct"

models["facts-14b-003"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-14b-003"].name = "facts-14b-003"
models["facts-14b-003"].base_model = "Qwen/Qwen2.5-14B-Instruct"
models["facts-14b-003"].config.scale_rewards = True
models["facts-14b-003"].config.trajectories_per_group = 12


models["facts-7b-001"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-7b-001"].name = "facts-7b-001"
models["facts-7b-001"].base_model = "Qwen/Qwen2.5-7B-Instruct"

models["facts-qwen3-4b-001"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-qwen3-4b-001"].name = "facts-qwen3-4b-001"
models["facts-qwen3-4b-001"].base_model = "Qwen/Qwen3-4B"

models["facts-qwen3-4b-002"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-qwen3-4b-002"].name = "facts-qwen3-4b-002"
models["facts-qwen3-4b-002"].base_model = "Qwen/Qwen3-4B"

models["facts-qwen3-4b-003"] = models["facts-14b-001"].model_copy(deep=True)
models["facts-qwen3-4b-003"].name = "facts-qwen3-4b-003"
models["facts-qwen3-4b-003"].base_model = "Qwen/Qwen3-4B"
models["facts-qwen3-4b-003"].config.groups_per_step = 8



from pydantic import BaseModel

from .. import dev, types
from ..preprocessing.pack import DiskPackedTensors


class Batch(BaseModel):
    disk_packed_tensors: DiskPackedTensors
    config: types.TrainConfig
    dev_config: dev.TrainConfig



import json
from copy import deepcopy
from typing import Iterable, List, Literal, Tuple

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from pydantic import create_model


def freeze_tool_schema(tool: dict, fixed_args: dict) -> ChatCompletionToolParam:
    """
    Return a clone of *tool* whose parameters schema permits *only* `fixed_args`.
    Each field is cast to typing.Literal[value] so Pydantic emits an
    enum-of-one in the JSON schema, which vLLM's `guided_json` accepts.
    """
    fields = {k: (Literal[v], ...) for k, v in fixed_args.items()}
    FrozenModel = create_model(
        f"{tool['function']['name'].title()}FrozenArgs",
        **fields,  # type: ignore
    )

    locked = deepcopy(tool)
    locked["function"]["parameters"] = FrozenModel.model_json_schema()
    return locked  # type: ignore


def get_guided_completion_params(
    completion: ChatCompletion,
    base_tools: Iterable[ChatCompletionToolParam] | None = None,
) -> Tuple[
    List[str] | None,
    ChatCompletionToolChoiceOptionParam | None,
    ChatCompletionToolParam | None,
]:
    """
    Given a completion from a teacher model, returns chat completion params that can be used to guide a student model's response.
    Useful for RL-based distillation.

    When guiding the student model's completion, remember to set `num_scheduler_steps` to 1.

    Args:
        completion: The completion of a teacher model
        base_tools: The base tools available to the teacher model

    Returns a tuple of (guided_choice, tool_choice, tool_params).
    """
    guided_choice, tool_choice, tool_params = None, None, None

    if (
        completion.choices[0].message.tool_calls
        and len(completion.choices[0].message.tool_calls) > 0
    ):
        tool_call = completion.choices[0].message.tool_calls[0]
        if not tool_call:
            raise ValueError("No tool call found in completion")
        if base_tools is None:
            raise ValueError("No base tools provided")
        tool_name = tool_call.function.name
        tool_choice = {
            "type": "function",  # ← must call it
            "function": {"name": tool_name},
        }
        chosen_tool = next(t for t in base_tools if t["function"]["name"] == tool_name)
        tool_params = [
            freeze_tool_schema(chosen_tool, json.loads(tool_call.function.arguments))  # type: ignore
        ]
    else:
        content = completion.choices[0].message.content
        guided_choice = [content]
    return (guided_choice, tool_choice, tool_params)  # type: ignore



import asyncio
from typing import Awaitable, Iterable

import tqdm

import art
from art.gather import GatherContext, set_gather_context, wrap_group_awaitable


async def gather_trajectory_groups_by_index(
    grouped_trajectory_awaitables: Iterable[Awaitable[tuple[art.Trajectory, ...]]],
    *,
    pbar_desc: str | None = "gather",
    pbar_total_completion_tokens: bool = True,
    max_exceptions: int | float = 0,
    trajectories_per_rollout: int = 1,
) -> list[art.TrajectoryGroup]:
    grouped_trajectory_awaitables = list(grouped_trajectory_awaitables)
    context = GatherContext(
        pbar=None,
        pbar_total_completion_tokens=pbar_total_completion_tokens,
        max_exceptions=max_exceptions,
    )

    with set_gather_context(context):
        future = asyncio.gather(
            *[wrap_group_awaitable(g) for g in grouped_trajectory_awaitables]
        )
        total = (
            sum(
                getattr(g, "_num_trajectories", 1)
                for g in grouped_trajectory_awaitables
            )
            * trajectories_per_rollout
        )
        context.pbar = tqdm.tqdm(desc=pbar_desc, total=total)
        all_trajectory_tuples = await future

    if context.pbar is not None:
        context.pbar.close()

    # Group trajectories by position in tuple
    num_groups = len(all_trajectory_tuples[0])
    grouped_trajectories: list[list[art.Trajectory]] = [[] for _ in range(num_groups)]

    for traj_tuple in all_trajectory_tuples:
        for i, traj in enumerate(traj_tuple):
            grouped_trajectories[i].append(traj)

    return [art.TrajectoryGroup(trajs) for trajs in grouped_trajectories]



from just_the_facts.rollout import FactsScenario

train_urls = [
    "https://www.nbcnews.com/select/shopping/best-dusters-rcna225656",
    "https://www.foxnews.com/us/us-military-raid-syria-eliminates-isis-leader-waiting-key-financier-officials",
    "https://www.bbc.com/news/articles/cgqnzy9ej0do?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/business/economy/trump-lisa-cook-federal-reserve-feud-flares-up-rcna226049",
    "https://www.foxnews.com/politics/mamdani-sidesteps-communism-question-says-his-campaign-about-distribution-wealth",
    "https://www.nbcnews.com/news/us-news/video-shows-colombian-tiktoker-tatiana-martinez-yanked-car-ice-arrest-rcna226038",
    "https://www.foxnews.com/politics/newsom-style-redistricting-efforts-critiqued-california-democrats-recently-july-statements-show",
    "https://www.nbcnews.com/weather/hurricanes/hurricane-erin-prompts-tropical-storm-warning-north-carolina-rcna226017",
    "https://www.cbsnews.com/news/google-pixel-10-product-launch-jimmy-fallon-gemini/",
    "https://www.cbsnews.com/news/trump-lisa-cook-federal-reserve-governor-bill-pulte/",
    "https://www.foxnews.com/politics/vance-says-national-guard-busting-a-washington-floats-mission-extension",
    "https://www.bbc.com/news/articles/c1ejvxne7elo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/world/ukraines-stolen-children-crisis-looms-large-nato-meets-russias-war",
    "https://www.bbc.com/news/articles/clyr7l0z9edo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cly4v3pllgko?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/c24zdel5j18o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/news/obituaries/tiktok-star-john-crawley-better-known-kingbeardx-dies-47-rcna226055",
    "https://www.cbsnews.com/news/ten-commandments-texas-abbott-law-blocked-federal-judge/",
    "https://www.bbc.com/news/articles/cd0d31r423ko?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/select/shopping/hand-soap-recall-rcna224873",
    "https://www.foxnews.com/us/family-torn-georgia-teen-accused-killing-parents-denied-bond-emotional-hearing",
    "https://www.foxnews.com/media/chris-matthews-tells-dems-abandon-all-other-crap-open-borders-trans-issues-focus-economy",
    "https://www.bbc.com/news/articles/c3r413l5n57o?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cj0yd0829m4o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/business/markets/trump-bought-bonds-companies-governments-rcna226024",
    "https://www.bbc.com/news/articles/c4g64qedynlo?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/politics/justice-department/judge-denies-doj-request-release-jeffrey-epstein-grand-jury-transcript-rcna226117",
    "https://www.cbsnews.com/news/hurricane-erin-track-path/",
    "https://www.foxnews.com/politics/duffys-dot-accuses-biden-buttigieg-inflating-air-traffic-controller-pipeline-juiced-numbers",
    "https://www.cbsnews.com/news/dhs-secretary-says-entire-southern-border-wall-to-be-painted-black-to-stop-people-from-climbing-it/",
    "https://www.bbc.com/news/articles/cy08309kxw6o?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/us-deportation-agreements-honduras-uganda/",
    "https://www.bbc.com/news/articles/c776kynn771o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/politics/vance-white-house-blast-crazy-communists-protesting-dc-clean-uo-terrorizing-locals-stupid-white-hippies",
    "https://www.foxnews.com/health/jeff-bezos-moms-death-spotlights-lewy-body-dementia-symptoms-risks",
    "https://www.bbc.com/news/articles/cy8jlwee0xjo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/politics/letitia-james-lawyer-blasts-ed-martins-bizarre-stunt-new-york-ags-house",
    "https://www.foxnews.com/politics/hillary-clinton-slapped-with-ethics-complaint-targeting-her-law-license-over-role-in-russiagate-scandal",
    "https://www.bbc.com/news/articles/c20662dxl88o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/us/second-federal-judge-blocks-justice-department-bid-release-jeffrey-epstein-grand-jury-files",
    "https://www.nbcnews.com/news/us-news/construction-companies-blamed-outbreak-legionnaires-rcna226104",
    "https://www.nbcnews.com/news/asian-america/fort-bliss-japanese-americans-internment-camp-immigrant-detention-rcna226044",
    "https://www.cbsnews.com/news/ernesto-barajas-killed-musician-drug-cartels-shot-dead-mexico/",
    "https://www.nbcnews.com/news/us-news/california-resident-catches-plague-likely-infected-flea-lake-tahoe-off-rcna226105",
    "https://www.cbsnews.com/news/abrupt-antarctic-climate-shifts-catastrophic-consequences-for-generations/",
    "https://www.foxnews.com/politics/fbi-arrests-woman-ten-most-wanted-fugitives-hiding-india-transports-us-prosecution",
    "https://www.foxnews.com/media/amanda-knox-how-one-misunderstood-text-sparked-suspicion-infamous-murder-case",
    "https://www.foxnews.com/sports/former-vikings-player-speaks-out-against-team-employing-male-cheerleaders",
    "https://www.foxnews.com/politics/illegal-trucker-deported-himself-ca-lawmaker-says-revealing-systemic-crisis-transportation",
    "https://www.bbc.com/news/articles/c939v4kw2l4o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/sports/tennis/carlos-alcaraz-emma-raducanu-us-open-first-round-doubles-rcna226089",
    "https://www.bbc.com/news/articles/cm2k1m56xgjo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cn92e52rpjxo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/media/white-house-rips-ex-white-stripes-singer-slamming-trump-oval-office-redecorating-gaudy",
    "https://www.nbcnews.com/business/markets/trump-bought-bonds-companies-governments-rcna226024",
    "https://www.bbc.com/news/articles/c30z944q42do?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/trump-doj-crime-victims-sanctuary-cities-lawsuit/",
    "https://www.foxnews.com/politics/desantis-state-line-wont-protect-criminal-illegals-sends-chuck-norris-deputy-nab-fugitive-trucker",
    "https://www.bbc.com/news/articles/cq587vpe46eo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/c5yer12rp8po?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/media/chuck-todd-argues-putin-made-huge-miscalculation-not-giving-trump-win-alaska-summit",
    "https://www.nbcnews.com/select/shopping/best-hot-sauces-tested-rcna206590",
    "https://www.nbcnews.com/select/shopping/new-notable-august-2025-rcna225665",
    "https://www.nbcnews.com/pop-culture/pop-culture-news/justin-bieber-impersonator-dupes-dj-ends-las-vegas-stage-rcna226041",
    "https://www.foxnews.com/politics/illegal-migrants-charged-rape-aggravated-sodomy-kidnapping-woman-kansas-ice",
    "https://www.nbcnews.com/tech/misinformation/doctors-say-medical-misinformation-gotten-worse-survey-finds-rcna225804",
    "https://www.cbsnews.com/news/us-tariffs-trump-inflation-prices/",
    "https://www.cbsnews.com/losangeles/news/menendez-brothers-california-parole-board-hearings/",
    "https://www.bbc.com/weather/articles/cwy5r7xwq8xo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/us-anti-american-views-immigration-benefits/",
    "https://www.nbcnews.com/politics/justice-department/judge-denies-doj-request-release-jeffrey-epstein-grand-jury-transcript-rcna226117",
    "https://www.nbcnews.com/weather/heat/dangerous-heat-wildfire-conditions-take-hold-west-rcna226067",
    "https://www.cbsnews.com/news/this-week-on-sunday-morning-august-17-2025/",
    "https://www.bbc.com/sport/rugby-union/articles/c9d0nwepedeo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/jeffrey-epstein-grand-jury-judge/",
    "https://www.bbc.com/sport/tennis/articles/c5ypke1n361o?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/powerball-jackpot-numbers-lottery-drawing-643-million/",
    "https://www.cbsnews.com/news/texas-house-special-session-redistricting-vote-today/",
    "https://www.bbc.com/news/articles/cj6yd4d5zj6o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/news/us-news/man-accused-stalking-family-unitedhealthcare-ceo-put-leave-rcna226028",
    "https://www.nbcnews.com/news/us-news/construction-companies-blamed-outbreak-legionnaires-rcna226104",
    "https://www.foxnews.com/health/five-deaths-reported-amid-bacterial-infection-outbreak-major-city",
    "https://www.cbsnews.com/news/trump-putin-zelenskyy-meeting-russia-ukraine/",
    "https://www.cbsnews.com/news/covid-19-shot-recommendations-american-academy-of-pediatrics-cdc/",
    "https://www.bbc.com/news/articles/cyvn9z3y78lo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/ps5-price-increase-sony/",
    "https://www.foxnews.com/politics/bostons-wu-fires-back-bondi-other-cities-slam-feds-over-sanctuary-warnings",
    "https://www.msnbc.com/top-stories/latest/joseph-humire-defense-department-pentagon-tren-de-aragua-rcna225929",
    "https://www.msnbc.com/top-stories/latest/texas-democrat-spends-night-house-floor-nicole-collier-rcna225907",
    "https://www.msnbc.com/top-stories/latest/democrats-senate-bid-sherrod-brown-ohio-jd-vance-rcna225866",
    "https://www.bbc.com/news/articles/c4gmjm8z47jo",
    "https://www.bbc.com/news/articles/cdrk267ldy7o",
    "https://www.bbc.com/news/articles/cm2vz9jg8jzo",
    "https://www.nbcnews.com/business/markets/nasdaq-tumbles-jackson-hole-jitters-hit-tech-stocks-rcna225890",
    "https://www.nbcnews.com/news/us-news/fake-labubu-dolls-knows-lafufus-are-serious-choking-hazards-kill-kids-rcna225691",
    "https://www.nbcnews.com/business/markets/nasdaq-tumbles-jackson-hole-jitters-hit-tech-stocks-rcna225890",
    "https://www.nbcnews.com/news/us-news/walmart-shrimp-may-exposed-radioactive-material-fda-says-rcna225903",
    "https://www.cbsnews.com/news/venezuela-nicolas-maduro-deploying-militia-us-threats/",
]


val_urls = [
    "https://www.foxnews.com/health/bindi-irwin-surgery-valley-fever-spike-fat-burning-treadmill-trend",
    "https://www.bbc.com/news/articles/cn85z0z3gz3o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/travel/hurricane-erin-prompts-beach-closures-emergency-declarations-across-multiple-states",
    "https://www.foxnews.com/politics/full-s-new-york-republican-accuses-state-dems-hypocrisy-redistricting-push",
    "https://www.bbc.com/news/articles/c5y3599gx4qo",
    "https://www.cbsnews.com/news/politics-may-have-spurred-august-purge-of-5-veteran-fbi-agents/",
    "https://www.foxnews.com/politics/schiff-launches-legal-defense-fund-response-claims-trump-weaponizing-justice-system",
    "https://www.msnbc.com/rachel-maddow-show/maddowblog/trump-rig-midterm-elections-redistricting-mail-in-voting-census-rcna225826",
]

train_scenarios = [FactsScenario(article_url=url) for url in train_urls]

val_scenarios = [FactsScenario(article_url=url) for url in val_urls]



import asyncio
from typing import Callable, TypeVar

import httpx
import sky
from sky.core import endpoints

T = TypeVar("T")


async def to_thread_typed(func: Callable[[], T]) -> T:
    return await asyncio.to_thread(func)


async def get_task_status(cluster_name: str, task_name: str) -> sky.JobStatus | None:
    job_queue = await to_thread_typed(
        lambda: sky.stream_and_get(sky.queue(cluster_name))
    )

    for job in job_queue:
        if job["job_name"] == task_name:
            return job["status"]
    return None


async def get_task_job_id(cluster_name: str, task_name: str) -> str | None:
    job_queue = await to_thread_typed(
        lambda: sky.stream_and_get(sky.queue(cluster_name))
    )

    for job in job_queue:
        if job["job_name"] == task_name:
            return job["job_id"]
    return None


async def is_task_created(cluster_name: str, task_name: str) -> bool:
    task_status = await get_task_status(cluster_name, task_name)
    if task_status is None:
        return False
    return task_status in (
        sky.JobStatus.INIT,
        sky.JobStatus.PENDING,
        sky.JobStatus.SETTING_UP,
        sky.JobStatus.RUNNING,
    )


# wait for task to start running
# checks every 5 seconds for 1 minute
async def wait_for_task_to_start(cluster_name: str, task_name: str) -> None:
    task_status = await get_task_status(cluster_name, task_name)

    num_checks = 0
    while num_checks < 12:
        task_status = await get_task_status(cluster_name, task_name)
        if task_status is None:
            raise ValueError(f"Task {task_name} not found in cluster {cluster_name}")
        if task_status == sky.JobStatus.RUNNING:
            return
        await asyncio.sleep(5)
        num_checks += 1

    raise ValueError(
        f"Task {task_name} in cluster {cluster_name} failed to start within 60s"
    )


async def wait_for_art_server_to_start(cluster_name: str) -> str:
    print(f"Waiting for art server task to start on cluster {cluster_name}...")
    await wait_for_task_to_start(cluster_name, "art_server")
    print(
        f"Art server started on cluster {cluster_name}. Waiting for it to be ready..."
    )

    base_url = await get_art_server_base_url(cluster_name)

    num_checks = 0
    client = httpx.AsyncClient(
        base_url=base_url,
        timeout=10,
    )
    while num_checks < 12:
        try:
            response = await client.get("/healthcheck")
            if response.status_code == 200:
                return base_url
        except Exception:
            pass
        await asyncio.sleep(5)
        num_checks += 1

    return base_url


async def get_art_server_base_url(cluster_name: str) -> str:
    art_endpoint = await to_thread_typed(
        lambda: endpoints(cluster=cluster_name, port=7999)[7999]
    )
    return f"http://{art_endpoint}"


async def get_vllm_base_url(cluster_name: str) -> str:
    vllm_endpoint = await to_thread_typed(
        lambda: endpoints(cluster=cluster_name, port=8000)[8000]
    )
    return f"http://{vllm_endpoint}/v1"



import asyncio
from typing import AsyncIterator, Awaitable, Iterable

from tqdm import auto as tqdm

from .gather import GatherContext, set_gather_context, wrap_group_awaitable
from .trajectories import TrajectoryGroup


async def trajectory_group_batches(
    groups: Iterable[Awaitable[TrajectoryGroup]],
    *,
    batch_size: int,
    max_batch_exceptions: int | float = 0,
    max_concurrent_batches: int = 4,
    skip_batches: int = 0,
    pbar_desc: str | None = "batches",
    pbar_total_completion_tokens: bool = True,
) -> AsyncIterator[list[TrajectoryGroup]]:
    unstarted = list(groups)[batch_size * skip_batches :]
    pending = set[asyncio.Task[TrajectoryGroup | None]]()
    batch = list[TrajectoryGroup]()
    context = GatherContext(
        pbar_total_completion_tokens=pbar_total_completion_tokens,
        max_exceptions=max_batch_exceptions,
        increment_pbar=False,
    )
    with set_gather_context(context):
        while unstarted or pending:
            if context.pbar is None:
                context.pbar = tqdm.tqdm(desc=pbar_desc, total=batch_size)
            while len(pending) < batch_size * max_concurrent_batches and unstarted:
                pending.add(asyncio.create_task(wrap_group_awaitable(unstarted.pop(0))))
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )
            context.pbar.update(len(done))
            batch.extend(g for task in done if (g := task.result()) is not None)
            if len(batch) >= batch_size:
                if context.pbar is not None:
                    context.pbar.close()
                    context.reset()
                yield batch[:batch_size]
                batch = batch[batch_size:]
        if batch:
            yield batch



"""Utilities for working with S3 checkpoints."""

import asyncio
from asyncio.subprocess import PIPE


async def get_latest_checkpoint_step_from_s3(
    model_name: str,
    project: str,
    s3_bucket: str | None = None,
    prefix: str | None = None,
) -> int | None:
    """
    Get the latest checkpoint step number from S3 without downloading files.

    Returns:
        The latest step number, or None if no checkpoints exist.
    """
    from .s3 import build_s3_path

    s3_path = build_s3_path(
        model_name=model_name,
        project=project,
        s3_bucket=s3_bucket,
        prefix=prefix,
    )

    # List checkpoint directories in S3
    cmd = ["aws", "s3", "ls", f"{s3_path}/checkpoints/"]

    process = await asyncio.create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        # No checkpoints found or error
        return None

    # Parse output to find checkpoint directories
    lines = stdout.decode().strip().split("\n")
    checkpoint_steps = []

    for line in lines:
        if line.strip():
            # AWS S3 ls output format: "PRE 0001/"
            parts = line.split()
            if len(parts) >= 2 and parts[0] == "PRE":
                dirname = parts[1].rstrip("/")
                if dirname.isdigit():
                    checkpoint_steps.append(int(dirname))

    return max(checkpoint_steps) if checkpoint_steps else None


async def get_checkpoint_step_not_after_from_s3(
    model_name: str,
    project: str,
    not_after_step: int,
    s3_bucket: str | None = None,
    prefix: str | None = None,
) -> int | None:
    """
    Get the latest checkpoint step number that is not after the specified step from S3.

    Args:
        not_after_step: Find the latest checkpoint <= this step.

    Returns:
        The step number, or None if no suitable checkpoint exists.
    """
    from .s3 import build_s3_path

    s3_path = build_s3_path(
        model_name=model_name,
        project=project,
        s3_bucket=s3_bucket,
        prefix=prefix,
    )

    # List checkpoint directories in S3
    cmd = ["aws", "s3", "ls", f"{s3_path}/checkpoints/"]

    process = await asyncio.create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        # No checkpoints found or error
        return None

    # Parse output to find checkpoint directories
    lines = stdout.decode().strip().split("\n")
    valid_steps = []

    for line in lines:
        if line.strip():
            # AWS S3 ls output format: "PRE 0001/"
            parts = line.split()
            if len(parts) >= 2 and parts[0] == "PRE":
                dirname = parts[1].rstrip("/")
                if dirname.isdigit():
                    step = int(dirname)
                    if step <= not_after_step:
                        valid_steps.append(step)

    return max(valid_steps) if valid_steps else None


async def migrate_s3_checkpoints_to_new_structure(
    model_name: str,
    project: str,
    s3_bucket: str | None = None,
    prefix: str | None = None,
    dry_run: bool = False,
) -> None:
    """
    Migrate existing checkpoints in S3 from the old structure to the new structure.

    Old: s3://bucket/prefix/project/models/model_name/0001/
    New: s3://bucket/prefix/project/models/model_name/checkpoints/0001/

    Args:
        model_name: The name of the model to migrate.
        project: The project name.
        s3_bucket: The S3 bucket. If None, uses BACKUP_BUCKET env var.
        prefix: Optional prefix for the S3 path.
        dry_run: If True, only print what would be done without making changes.
    """
    import os

    from .s3 import build_s3_path

    if s3_bucket is None:
        s3_bucket = os.environ.get("BACKUP_BUCKET")
        if not s3_bucket:
            raise ValueError(
                "BACKUP_BUCKET environment variable not set and no bucket provided"
            )

    s3_path = build_s3_path(
        model_name=model_name,
        project=project,
        s3_bucket=s3_bucket,
        prefix=prefix,
    )

    print(f"Checking for checkpoints to migrate in {s3_path}")

    # List all directories in the model path
    cmd = ["aws", "s3", "ls", f"{s3_path}/"]
    process = await asyncio.create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        print(f"Error listing S3 path: {stderr.decode()}")
        return

    # Parse output to find checkpoint directories
    lines = stdout.decode().strip().split("\n")
    checkpoint_dirs = []

    for line in lines:
        if line.strip():
            # AWS S3 ls output format: "PRE 0001/"
            parts = line.split()
            if len(parts) >= 2 and parts[0] == "PRE":
                dirname = parts[1].rstrip("/")
                # Check if it's a 4-digit checkpoint directory (old format)
                if dirname.isdigit() and len(dirname) == 4:
                    checkpoint_dirs.append(dirname)

    if not checkpoint_dirs:
        print("No checkpoints found in old format to migrate")
        return

    print(
        f"Found {len(checkpoint_dirs)} checkpoint(s) to migrate: {', '.join(checkpoint_dirs)}"
    )

    if dry_run:
        print("DRY RUN: Would migrate the following checkpoints:")
        for checkpoint in checkpoint_dirs:
            print(f"  {s3_path}/{checkpoint}/ -> {s3_path}/checkpoints/{checkpoint}/")
        return

    # Perform migration
    migrated_count = 0
    for checkpoint in checkpoint_dirs:
        old_path = f"{s3_path}/{checkpoint}/"
        new_path = f"{s3_path}/checkpoints/{checkpoint}/"

        print(f"Migrating checkpoint {checkpoint}...")

        # Check if already exists in new location
        check_cmd = ["aws", "s3", "ls", new_path]
        check_process = await asyncio.create_subprocess_exec(
            *check_cmd, stdout=PIPE, stderr=PIPE
        )
        check_stdout, _ = await check_process.communicate()

        if check_process.returncode == 0 and check_stdout.decode().strip():
            print(f"  Checkpoint {checkpoint} already exists in new location, skipping")
            continue

        # Copy checkpoint to new location (using sync to preserve structure)
        sync_cmd = ["aws", "s3", "sync", old_path, new_path]
        sync_process = await asyncio.create_subprocess_exec(
            *sync_cmd, stdout=PIPE, stderr=PIPE
        )
        _, sync_stderr = await sync_process.communicate()

        if sync_process.returncode != 0:
            print(f"  Error copying checkpoint {checkpoint}: {sync_stderr.decode()}")
            continue

        # Verify copy was successful by checking if files exist in new location
        verify_cmd = ["aws", "s3", "ls", new_path, "--recursive"]
        verify_process = await asyncio.create_subprocess_exec(
            *verify_cmd, stdout=PIPE, stderr=PIPE
        )
        verify_stdout, _ = await verify_process.communicate()

        if verify_process.returncode != 0 or not verify_stdout.decode().strip():
            print(
                f"  Error: Checkpoint {checkpoint} not found in new location after copy"
            )
            continue

        # Remove old checkpoint directory
        rm_cmd = ["aws", "s3", "rm", old_path, "--recursive"]
        rm_process = await asyncio.create_subprocess_exec(
            *rm_cmd, stdout=PIPE, stderr=PIPE
        )
        _, rm_stderr = await rm_process.communicate()

        if rm_process.returncode != 0:
            print(
                f"  Warning: Failed to remove old checkpoint {checkpoint}: {rm_stderr.decode()}"
            )
            print(
                "  Checkpoint was successfully copied to new location but old files remain"
            )
        else:
            print(f"  Successfully migrated checkpoint {checkpoint}")
            migrated_count += 1

    print(f"\nMigration complete. Successfully migrated {migrated_count} checkpoint(s)")



# Import all utilities to maintain the same interface
from .format_message import format_message
from .get_model_step import get_model_step
from .iterate_dataset import iterate_dataset
from .limit_concurrency import limit_concurrency
from .log_http_errors import log_http_errors
from .retry import retry

__all__ = [
    "format_message",
    "retry",
    "iterate_dataset",
    "limit_concurrency",
    "log_http_errors",
    "get_model_step",
]



import json
import socket
from typing import Any, AsyncIterator

import pydantic
import typer
import uvicorn
from dotenv import load_dotenv
from fastapi import Body, FastAPI, Request
from fastapi.responses import JSONResponse, StreamingResponse

from . import dev
from .errors import ARTError
from .local import LocalBackend
from .model import Model, TrainableModel
from .trajectories import TrajectoryGroup
from .types import TrainConfig
from .utils.deploy_model import LoRADeploymentProvider

load_dotenv()

app = typer.Typer()


@app.command()
def run(host: str = "0.0.0.0", port: int = 7999) -> None:
    """Run the ART CLI."""

    # check if port is available
    def is_port_available(port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(("localhost", port)) != 0

    if not is_port_available(port):
        print(
            f"Port {port} is already in use, possibly because the ART server is already running."
        )
        return

    # Reset the custom __new__ and __init__ methods for TrajectoryGroup
    def __new__(cls, *args: Any, **kwargs: Any) -> TrajectoryGroup:
        return pydantic.BaseModel.__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        return pydantic.BaseModel.__init__(self, *args, **kwargs)

    TrajectoryGroup.__new__ = __new__  # type: ignore
    TrajectoryGroup.__init__ = __init__

    backend = LocalBackend()
    app = FastAPI()

    # Add exception handler for ARTError
    @app.exception_handler(ARTError)
    async def art_error_handler(request: Request, exc: ARTError):
        return JSONResponse(status_code=exc.status_code, content={"detail": exc.detail})

    app.get("/healthcheck")(lambda: {"status": "ok"})
    app.post("/close")(backend.close)
    app.post("/register")(backend.register)
    app.post("/_get_step")(backend._get_step)
    app.post("/_delete_checkpoints")(backend._delete_checkpoints)

    @app.post("/_prepare_backend_for_training")
    async def _prepare_backend_for_training(
        model: TrainableModel,
        config: dev.OpenAIServerConfig | None = Body(None),
    ):
        return await backend._prepare_backend_for_training(model, config)

    @app.post("/_log")
    async def _log(
        model: Model,
        trajectory_groups: list[TrajectoryGroup],
        split: str = Body("val"),
    ):
        await backend._log(model, trajectory_groups, split)

    @app.post("/_train_model")
    async def _train_model(
        model: TrainableModel,
        trajectory_groups: list[TrajectoryGroup],
        config: TrainConfig,
        dev_config: dev.TrainConfig,
        verbose: bool = Body(False),
    ) -> StreamingResponse:
        async def stream() -> AsyncIterator[str]:
            async for result in backend._train_model(
                model, trajectory_groups, config, dev_config, verbose
            ):
                yield json.dumps(result) + "\n"

        return StreamingResponse(stream())

    # Wrap in function with Body(...) to ensure FastAPI correctly interprets
    # all parameters as body parameters
    @app.post("/_experimental_pull_from_s3")
    async def _experimental_pull_from_s3(
        model: Model = Body(...),
        s3_bucket: str | None = Body(None),
        prefix: str | None = Body(None),
        verbose: bool = Body(False),
        delete: bool = Body(False),
    ):
        await backend._experimental_pull_from_s3(
            model=model,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            delete=delete,
        )

    @app.post("/_experimental_push_to_s3")
    async def _experimental_push_to_s3(
        model: Model = Body(...),
        s3_bucket: str | None = Body(None),
        prefix: str | None = Body(None),
        verbose: bool = Body(False),
        delete: bool = Body(False),
    ):
        await backend._experimental_push_to_s3(
            model=model,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            delete=delete,
        )

    @app.post("/_experimental_deploy")
    async def _experimental_deploy(
        deploy_to: LoRADeploymentProvider = Body(...),
        model: TrainableModel = Body(...),
        step: int | None = Body(None),
        s3_bucket: str | None = Body(None),
        prefix: str | None = Body(None),
        verbose: bool = Body(False),
        pull_s3: bool = Body(True),
        wait_for_completion: bool = Body(True),
    ):
        return await backend._experimental_deploy(
            deploy_to=deploy_to,
            model=model,
            step=step,
            s3_bucket=s3_bucket,
            prefix=prefix,
            verbose=verbose,
            pull_s3=pull_s3,
            wait_for_completion=wait_for_completion,
        )

    uvicorn.run(app, host=host, port=port, loop="asyncio")



from mcp import types


def get_content_text(result: types.CallToolResult) -> str:
    # Extract text content from MCP result
    if hasattr(result, "content") and result.content:
        if isinstance(result.content, list):
            # Handle list of content items
            content_text = ""
            for item in result.content:
                if isinstance(item, types.TextContent):
                    content_text += item.text
                else:
                    content_text += str(item)
        elif isinstance(result.content[0], types.TextContent):
            content_text = result.content[0].text
        else:
            content_text = str(result.content)
    else:
        content_text = str(result)

    return content_text



import argparse
import ast
import re
import sys
from pathlib import Path
from typing import Any, Dict

import nbformat
import pytest
from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError

NOTEBOOKS = [
    {
        "path": "../examples/temporal_clue/temporal-clue.ipynb",
        "variables": {
            "TRAINING_STEPS": 1,
        },
    },
    {
        "path": "../examples/prisoners-dilemma.ipynb",
        "variables": {
            "TRAINING_STEPS": 1,
            "PRISONERS_DILEMMA_ROUNDS": 10,
        },
    },
    {
        "path": "../examples/art-e.ipynb",
        "variables": {
            "training_config": {
                "groups_per_step": 2,
                "num_epochs": 1,
                "rollouts_per_group": 4,
                "learning_rate": 1e-5,
                "max_steps": 1,
            },
        },
    },
    {
        "path": "../examples/rock-paper-tool-use.ipynb",
        "variables": {"TRAINING_STEPS": 1},
    },
]


def make_patch_source() -> str:
    """
    This is a patch to the art.TrainableModel class to change project name logging these runs as test runs.
    """
    project = "Tester"
    return (
        "try:\n"
        "    import art as _art\n"
        "    import art.model as _art_model\n"
        "except Exception:\n"
        "    pass\n"
        "else:\n"
        "    _orig_tm_init = _art_model.TrainableModel.__init__\n"
        "    def _patched_tm_init(self, *args, **kwargs):\n"
        f"        kwargs['project'] = {project!r}\n"
        "        result = _orig_tm_init(self, *args, **kwargs)\n"
        "        return result\n"
        "    _art_model.TrainableModel.__init__ = _patched_tm_init\n"
        "    _art.TrainableModel = _art_model.TrainableModel\n"
    )


def make_variable_override_source(variables: Dict[str, Any]) -> str:
    """
    Create source code to override variables in the notebook.
    """
    if not variables:
        return ""

    lines = ["# Variable overrides for testing"]
    for var_name, var_value in variables.items():
        lines.append(f"{var_name} = {var_value!r}")

    return "\n".join(lines)


def _override_variables_in_notebook(nb, variables: Dict[str, Any]) -> list[str]:
    """Replace top-level assignments inside code cells.

    Replaces entire assignment statements (including multi-line values) using AST
    ranges to avoid leaving trailing lines that can break indentation.

    Returns the list of variable names that were NOT found and replaced.
    """
    if not variables:
        return []

    replaced: set[str] = set()

    for cell in nb.cells:
        if getattr(cell, "cell_type", "") != "code":
            continue
        source: str = getattr(cell, "source", "") or ""
        if not source:
            continue

        # Try AST-based replacement first
        try:
            tree = ast.parse(source)
        except SyntaxError:
            tree = None

        if tree is not None:
            lines = source.splitlines(keepends=True)
            pending_edits: list[tuple[int, int, str]] = []

            for node in getattr(tree, "body", []) or []:
                if isinstance(node, (ast.Assign, ast.AnnAssign)):
                    targets = (
                        node.targets if isinstance(node, ast.Assign) else [node.target]
                    )
                    for t in targets:
                        if isinstance(t, ast.Name):
                            var_name = t.id
                            if var_name in variables and var_name not in replaced:
                                start_ln = getattr(node, "lineno", 1) - 1
                                end_ln = (
                                    getattr(
                                        node, "end_lineno", getattr(node, "lineno", 1)
                                    )
                                    - 1
                                )
                                # Preserve exact indentation used at the start line
                                indent_match = re.match(r"[ \t]*", lines[start_ln])
                                indent = indent_match.group(0) if indent_match else ""
                                replacement = (
                                    f"{indent}{var_name} = {variables[var_name]!r}\n"
                                )
                                pending_edits.append((start_ln, end_ln, replacement))
                                replaced.add(var_name)
                                break

            if pending_edits:
                # Apply edits from bottom to top to keep indices valid
                pending_edits.sort(key=lambda e: e[0], reverse=True)
                for start_ln, end_ln, replacement in pending_edits:
                    lines[start_ln : end_ln + 1] = [replacement]
                cell.source = "".join(lines)
                continue

        # Fallback to single-line regex replacement if AST failed or found nothing
        for var_name, var_value in variables.items():
            if var_name in replaced:
                continue
            pattern = re.compile(
                rf"^([ \t]*){re.escape(var_name)}\s*=\s*.*$", re.MULTILINE
            )

            def _sub(m):
                replaced.add(var_name)
                return f"{m.group(1)}{var_name} = {var_value!r}"

            source_new, n = pattern.subn(_sub, source, count=1)
            if n:
                source = source_new
        cell.source = source

    missing = [name for name in variables.keys() if name not in replaced]
    return missing


class _NotebookPlugin:
    def __init__(self, notebook_configs: list[dict]) -> None:
        self.notebook_configs = notebook_configs

    def pytest_generate_tests(self, metafunc) -> None:
        if "notebook_config" in metafunc.fixturenames:
            metafunc.parametrize("notebook_config", self.notebook_configs)


def test_notebook_execution(notebook_config: dict) -> None:
    notebook_path = notebook_config["path"]
    variables = notebook_config.get("variables", {})

    p = Path(notebook_path).resolve()
    if not (p.is_file() and p.suffix == ".ipynb"):
        pytest.skip(f"Notebook not found or invalid: {p}")

    nb = nbformat.read(p, as_version=4)

    # Replace variables directly inside existing cells first
    missing_variables: list[str] = _override_variables_in_notebook(nb, variables)

    # Insert the patch source at the beginning
    nb.cells.insert(0, nbformat.v4.new_code_cell(source=make_patch_source()))

    # For variables not found in the notebook, insert a small override cell
    if missing_variables:
        override_source = make_variable_override_source(
            {k: variables[k] for k in missing_variables}
        )
        nb.cells.insert(1, nbformat.v4.new_code_cell(source=override_source))

    try:
        NotebookClient(nb).execute(cwd=str(p.parent))
    except CellExecutionError as e:
        pytest.fail(str(e))
    # Echo cell outputs to stdout/stderr so prints are visible in console
    for cell in nb.cells:
        if getattr(cell, "cell_type", None) != "code":
            continue
        for output in getattr(cell, "outputs", []) or []:
            output_type = output.get("output_type")
            if output_type == "stream":
                text = output.get("text", "")
                name = output.get("name", "stdout")
                if name == "stderr":
                    print(text, end="", file=sys.stderr)
                else:
                    print(text, end="")
            elif output_type in ("execute_result", "display_data"):
                data = output.get("data", {}) or {}
                if "text/plain" in data:
                    print(data["text/plain"])
            elif output_type == "error":
                traceback_lines = output.get("traceback") or []
                if traceback_lines:
                    print("\n".join(traceback_lines), file=sys.stderr)

    # Print success message after test completes successfully
    notebook_name = Path(notebook_path).name
    print(f"\n✅ Test completed successfully: {notebook_name}")
    print("-" * 50)


def parse_indexes(indexes_str: str, max_index: int) -> list[int]:
    """
    Parse index specification string into list of valid indexes.

    Supports:
    - Single indexes: "0", "2"
    - Comma-separated: "0,2,3"
    - Ranges: "1-3" (inclusive)
    - Mixed: "0,2-4,6"

    Args:
        indexes_str: String specification of indexes
        max_index: Maximum valid index (exclusive)

    Returns:
        List of valid, unique indexes in ascending order
    """
    indexes = set()

    try:
        for part in indexes_str.split(","):
            part = part.strip()
            if not part:
                continue

            if "-" in part:
                # Handle range like "1-3"
                start_str, end_str = part.split("-", 1)
                start = int(start_str.strip())
                end = int(end_str.strip())

                if start > end:
                    print(f"Warning: Invalid range '{part}' - start > end")
                    continue

                for i in range(start, end + 1):
                    if 0 <= i < max_index:
                        indexes.add(i)
                    else:
                        print(f"Warning: Index {i} is out of range (0-{max_index - 1})")
            else:
                # Handle single index
                index = int(part)
                if 0 <= index < max_index:
                    indexes.add(index)
                else:
                    print(f"Warning: Index {index} is out of range (0-{max_index - 1})")

    except ValueError as e:
        print(f"Error parsing indexes '{indexes_str}': {e}")
        return []

    return sorted(list(indexes))


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Run integration tests for notebooks",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=f"""
Available notebooks (by index):
{chr(10).join(f"  {i}: {config['path']}" for i, config in enumerate(NOTEBOOKS))}

Examples:
  uv run integration.py              # Run all notebooks
  uv run integration.py -n 0         # Run only notebook 0
  uv run integration.py -n 0,2       # Run notebooks 0 and 2
  uv run integration.py -n 1-3       # Run notebooks 1, 2, and 3
""",
    )
    parser.add_argument(
        "--notebooks",
        "-n",
        type=str,
        help="Comma-separated list of notebook indexes to run (e.g., '0,2' or '1-3'). If not specified, all notebooks will be run.",
    )

    args = parser.parse_args()

    here = Path(__file__).parent.resolve()

    # Determine which notebooks to run
    if args.notebooks:
        selected_indexes = parse_indexes(args.notebooks, len(NOTEBOOKS))
        if not selected_indexes:
            print("Error: No valid notebook indexes specified")
            return 1
        selected_configs = [NOTEBOOKS[i] for i in selected_indexes]
        print(f"Running notebooks at indexes: {sorted(selected_indexes)}")
    else:
        selected_configs = NOTEBOOKS
        print("Running all notebooks")

    # Process notebook configurations
    processed_configs = []
    for config in selected_configs:
        if isinstance(config, str):
            # Handle legacy string format
            processed_config = {"path": config, "variables": {}}
        else:
            processed_config = config.copy()

        # Resolve path relative to this file
        p = (here / processed_config["path"]).resolve()
        if not p.exists():
            print(f"Warning: notebook not found: {p}")
        processed_config["path"] = str(p)
        processed_configs.append(processed_config)

    # Invoke pytest programmatically, injecting notebook params via a plugin
    # -s: do not capture stdout so notebook output is visible
    # --maxfail=1: stop on first failure to terminate the whole training run
    pytest_args = [
        "-s",
        "--maxfail=1",
        str(here / "integration.py"),
    ]
    result_code = pytest.main(
        args=pytest_args, plugins=[_NotebookPlugin(processed_configs)]
    )
    return int(result_code)


if __name__ == "__main__":
    sys.exit(main())



import os

from art.local import LocalBackend
from art.model import Model as ArtModel


async def pull_model_trajectories(model: ArtModel) -> None:
    """Pull trajectory checkpoints for *model* from the configured S3 bucket.

    This is a lightweight helper that mirrors the S3-sync logic used inside
    ``art_e.train`` but without performing any training.  It can be invoked from
    notebooks or other scripts to ensure that the local ART project directory
    contains all trajectory files for subsequent evaluation / analysis.

    Parameters
    ----------
    model : art.Model
        Any ART model instance (trainable or not) that should be synchronised.

    Environment
    -----------
    BACKUP_BUCKET : str
        Name of the S3 bucket that stores model artefacts. The variable is
        loaded from the current environment (``dotenv`` is consulted so that
        values from a local *.env* file are respected).
    """

    bucket = os.getenv("BACKUP_BUCKET")
    if bucket is None:
        raise EnvironmentError(
            "Environment variable BACKUP_BUCKET is required but was not found."
        )

    # Use the LocalBackend context manager to work with the on-disk artefacts.
    with LocalBackend() as backend:
        print(
            f"Pulling trajectories for model '{model.name}' from S3 bucket '{bucket}'…",
            flush=True,
        )

        await backend._experimental_pull_from_s3(
            model,
            s3_bucket=bucket,
            verbose=True,
            exclude=["checkpoints", "logs"],
        )

        print("Finished pulling trajectories.", flush=True)



import warnings

# Suppress pydantic warnings at module level
warnings.filterwarnings("ignore", category=DeprecationWarning, module="pydantic")

import litellm
import litellm.litellm_core_utils.streaming_handler
import litellm.types.utils
import pytest
import pytest_asyncio
from aiohttp import web
from openai import AsyncOpenAI
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam

import art
from art.utils.litellm import convert_litellm_choice_to_openai

mock_response = {
    "id": "chatcmpl-293ce9f37dba40e5be39448acaf6fb49",
    "choices": [
        {
            "finish_reason": "stop",
            "index": 0,
            "logprobs": {
                "content": [
                    {
                        "token": "token_id:9707",
                        "bytes": [72, 101, 108, 108, 111],
                        "logprob": -0.0017243054462596774,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:0",
                        "bytes": [33],
                        "logprob": -0.007611795328557491,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:2585",
                        "bytes": [32, 72, 111, 119],
                        "logprob": -0.03061593696475029,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:646",
                        "bytes": [32, 99, 97, 110],
                        "logprob": -1.1920858014491387e-05,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:358",
                        "bytes": [32, 73],
                        "logprob": -2.3841855067985307e-07,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:7789",
                        "bytes": [32, 97, 115, 115, 105, 115, 116],
                        "logprob": -0.020548323169350624,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:498",
                        "bytes": [32, 121, 111, 117],
                        "logprob": 0.0,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:3351",
                        "bytes": [32, 116, 111, 100, 97, 121],
                        "logprob": -4.410734163684538e-06,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:30",
                        "bytes": [63],
                        "logprob": -2.3841855067985307e-07,
                        "top_logprobs": [],
                    },
                    {
                        "token": "token_id:151645",
                        "bytes": [],
                        "logprob": -0.0083366259932518,
                        "top_logprobs": [],
                    },
                ],
                "refusal": None,
            },
            "message": {
                "content": "Hello! How can I assist you today?",
                "refusal": None,
                "role": "assistant",
                "annotations": None,
                "audio": None,
                "function_call": None,
                "tool_calls": [],
                "reasoning_content": None,
            },
            "stop_reason": None,
        }
    ],
    "created": 1755801745,
    "model": "test",
    "object": "chat.completion",
    "service_tier": None,
    "system_fingerprint": None,
    "usage": {
        "completion_tokens": 10,
        "prompt_tokens": 31,
        "total_tokens": 41,
        "completion_tokens_details": None,
        "prompt_tokens_details": None,
    },
    "prompt_logprobs": None,
    "kv_transfer_params": None,
}
mock_stream_response = b"""data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"role":"assistant","content":""},"logprobs":null,"finish_reason":null}]}

data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"role":"assistant","content":""},"logprobs":null,"finish_reason":null}]}

data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"tool_calls":[{"id":"chatcmpl-tool-29e663261e524fcfa2162f4f3d76a7f0","type":"function","index":0,"function":{"name":"get_current_weather","arguments":"{"}}]},"logprobs":{"content":[{"token":"token_id:314","logprob":-0.00015293381875380874,"bytes":[32,123],"top_logprobs":[]}]},"finish_reason":null}]}

data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"tool_calls":[{"id":"chatcmpl-tool-29e663261e524fcfa2162f4f3d76a7f0","type":"function","index":0,"function":{"name":"get_current_weather","arguments":"{"}}]},"logprobs":{"content":[{"token":"token_id:314","logprob":-0.00015293381875380874,"bytes":[32,123],"top_logprobs":[]}]},"finish_reason":null}]}

data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"name":null,"arguments":"}"}}]},"logprobs":{"content":[{"token":"token_id:3417","logprob":-3.576278118089249e-7,"bytes":[125,125],"top_logprobs":[]}]},"finish_reason":null}]}

data: {"id":"chatcmpl-aa0d1e3261414f53acafc2f8e53bf9d6","object":"chat.completion.chunk","created":1755831263,"model":"test","choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"name":null,"arguments":"}"}}]},"logprobs":{"content":[{"token":"token_id:3417","logprob":-3.576278118089249e-7,"bytes":[125,125],"top_logprobs":[]}]},"finish_reason":null}]}

data: [DONE]

data: [DONE]

"""
mock_stream_choice = Choice(
    **{
        "finish_reason": "stop",
        "index": 0,
        "logprobs": {
            "content": [
                {
                    "token": "token_id:314",
                    "bytes": [32, 123],
                    "logprob": -0.00015293381875380874,
                    "top_logprobs": [],
                },
                {
                    "token": "token_id:314",
                    "bytes": [32, 123],
                    "logprob": -0.00015293381875380874,
                    "top_logprobs": [],
                },
                {
                    "token": "token_id:3417",
                    "bytes": [125, 125],
                    "logprob": -3.576278118089249e-07,
                    "top_logprobs": [],
                },
                {
                    "token": "token_id:3417",
                    "bytes": [125, 125],
                    "logprob": -3.576278118089249e-07,
                    "top_logprobs": [],
                },
            ],
            "refusal": None,
        },
        "message": {
            "content": None,
            "refusal": None,
            "role": "assistant",
            "annotations": None,
            "audio": None,
            "function_call": None,
            "tool_calls": [
                {
                    "id": "chatcmpl-tool-29e663261e524fcfa2162f4f3d76a7f0",
                    "function": {"arguments": "{{}}", "name": "get_current_weather"},
                    "type": "function",
                }
            ],
        },
    }
)


@pytest_asyncio.fixture
async def test_server():
    """Start a test server for the module."""

    async def handler(request: web.Request) -> web.Response:
        body = await request.json()
        if body.get("stream", False):
            return web.Response(body=mock_stream_response)
        return web.json_response(mock_response)

    app = web.Application()
    app.router.add_route("POST", "/v1/chat/completions", handler)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, "localhost", 8888)
    await site.start()
    print(f"Test server started on http://localhost:8888")

    yield  # Tests run here

    print("Cleaning up test server...")
    await runner.cleanup()


async def test_auto_trajectory(test_server: None) -> None:
    message: ChatCompletionMessageParam = {"role": "user", "content": "Hi!"}
    tools: list[ChatCompletionToolParam] = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            },
        }
    ]

    async def say_hi() -> str | None:
        """A method that says hi to an assistant and returns the response."""
        client = AsyncOpenAI(base_url="http://localhost:8888/v1", api_key="default")
        chat_completion = await client.chat.completions.create(
            model="test",
            messages=[message],
            tools=tools,
        )
        # test a follow up message
        chat_completion = await client.chat.completions.create(
            model="test",
            messages=[
                message,
                {
                    "role": "assistant",
                    "content": chat_completion.choices[0].message.content,
                },
                message,
            ],
            tools=tools,
        )
        # and another call without tools (should create a new history)
        chat_completion = await client.chat.completions.create(
            model="test",
            messages=[
                message,
                {
                    "role": "assistant",
                    "content": chat_completion.choices[0].message.content,
                },
                message,
                {
                    "role": "assistant",
                    "content": chat_completion.choices[0].message.content,
                },
                message,
            ],
        )
        # and another call with tools, but limited messages (should create another history)
        chat_completion = await client.chat.completions.create(
            model="test",
            messages=[message],
            tools=tools,
        )
        # and another call with tool_choice="required" & stream=True
        async for _ in await client.chat.completions.create(
            model="test",
            messages=[message],
            tool_choice="required",
            tools=tools,
            stream=True,
        ):
            pass
        # Add ART support with a couple lines of optional code
        if trajectory := art.auto_trajectory():
            trajectory.reward = 1.0
        return chat_completion.choices[0].message.content

    # Use the capture_auto_trajectory utility to capture a trajectory automatically
    trajectory = await art.capture_auto_trajectory(say_hi())
    assert trajectory.messages_and_choices == [
        message,
        Choice(**mock_response["choices"][0]),
        message,
        Choice(**mock_response["choices"][0]),
    ]
    assert trajectory.tools == tools
    assert trajectory.additional_histories[0].messages_and_choices == [
        message,
        {
            "content": "Hello! How can I assist you today?",
            "role": "assistant",
        },
        message,
        {
            "content": "Hello! How can I assist you today?",
            "role": "assistant",
        },
        message,
        Choice(**mock_response["choices"][0]),
    ]
    assert trajectory.additional_histories[0].tools is None
    assert trajectory.additional_histories[1].messages_and_choices == [
        message,
        Choice(**mock_response["choices"][0]),
    ]
    assert trajectory.additional_histories[1].tools == tools
    assert trajectory.additional_histories[2].messages_and_choices == [
        message,
        mock_stream_choice,
    ]
    assert trajectory.additional_histories[2].tools == tools


@pytest.mark.filterwarnings("ignore::UserWarning:pydantic")
async def test_litellm_auto_trajectory(test_server: None) -> None:
    message: ChatCompletionMessageParam = {"role": "user", "content": "Hi!"}
    tools: list[ChatCompletionToolParam] = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            },
        }
    ]

    async def say_hi() -> str | None:
        """A method that says hi to an assistant and returns the response."""
        response = await litellm.acompletion(
            model="openai/test",
            messages=[message],
            tools=tools,
            base_url="http://localhost:8888/v1",
            api_key="default",
        )
        assert isinstance(response, litellm.types.utils.ModelResponse)
        choice = convert_litellm_choice_to_openai(response.choices[0])
        # follow up message
        response = await litellm.acompletion(
            model="openai/test",
            messages=[
                message,
                {"role": "assistant", "content": choice.message.content},
                message,
            ],
            tools=tools,
            base_url="http://localhost:8888/v1",
            api_key="default",
        )
        assert isinstance(response, litellm.types.utils.ModelResponse)
        choice = convert_litellm_choice_to_openai(response.choices[0])
        # another call with tool_choice="required" & stream=True
        stream = await litellm.acompletion(
            model="openai/test",
            messages=[message],
            tool_choice="required",
            tools=tools,
            stream=True,
            base_url="http://localhost:8888/v1",
            api_key="default",
        )
        assert isinstance(
            stream, litellm.litellm_core_utils.streaming_handler.CustomStreamWrapper
        )
        async for _ in stream:
            pass
        # Add ART support with a couple lines of optional code
        if trajectory := art.auto_trajectory():
            trajectory.reward = 1.0
        return choice.message.content

    # Use the capture_auto_trajectory utility to capture a trajectory automatically
    trajectory = await art.capture_auto_trajectory(say_hi())
    assert trajectory.messages_and_choices == [
        message,
        Choice(**mock_response["choices"][0]),
        message,
        Choice(**mock_response["choices"][0]),
    ]
    assert trajectory.additional_histories[0].messages_and_choices == [
        message,
        mock_stream_choice,
    ]
    assert trajectory.additional_histories[0].tools == tools



import os

from art.model import Model
from art.utils.get_repo_root_path import get_repo_root_path


def get_default_art_path() -> str:
    root_path = get_repo_root_path()
    return os.path.join(root_path, ".art")


def get_models_dir(project_name: str, art_path: str | None = None) -> str:
    if art_path is None:
        art_path = get_default_art_path()
    return f"{art_path}/{project_name}/models"


def get_model_dir(model: Model, art_path: str | None = None) -> str:
    if art_path is None:
        art_path = get_default_art_path()
    return f"{art_path}/{model.project}/models/{model.name}"


def get_output_dir_from_model_properties(
    project: str, name: str, art_path: str | None = None
) -> str:
    if art_path is None:
        art_path = get_default_art_path()
    return f"{art_path}/{project}/models/{name}"


def get_step_checkpoint_dir(model_output_dir: str, step: int) -> str:
    return f"{model_output_dir}/checkpoints/{step:04d}"


def get_trajectories_dir(model_output_dir: str) -> str:
    return f"{model_output_dir}/trajectories"


def get_trajectories_split_dir(model_output_dir: str, split: str) -> str:
    return f"{model_output_dir}/trajectories/{split}"
